"""
.. _ArrayTransforms
:mod:`utils` -- Array Transforms
================================

.. module:: ArrayTransforms
   :platform: Unix, Windows
   :synopsis: Provides utilities for array shape and type manipulation.
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
__all__ = ['interior_vertex_func', 'CenterTime', 'CenterLay', 'CenterRow',
           'CenterCol', 'CenterRowCol', 'CenterTimeRowCol', 'CenterCMAQWind',
           'CenterCAMxWind', 'CenterCAMxU', 'CenterCAMxV', 'BoundToDiff',
           'CAMxHeightToDepth', 'ConvertCAMxTime']

import unittest
from numpy import mean, array, sum, arange, zeros, newaxis
from warnings import warn
from PseudoNetCDF.sci_var import PseudoNetCDFVariable


def interior_vertex_func(a, dims=(-1, -2), func=sum):
    if func not in (sum, mean):
        warn("Function has only been tested with sum and mean")

    if len(dims) not in (1, 2, 3, 4):
        warn("Function has only been tested with up to 4-D")

    full_shape = array([slice(None)] * len(a.shape))
    out_array = array(a, copy=True)

    for dim in dims:
        zero_to_minus_one = full_shape.copy()
        one_to_end = full_shape.copy()
        zero_to_minus_one[dim] = slice(0, -1)
        one_to_end[dim] = slice(1, None)
        out_array = func(array([out_array[tuple(zero_to_minus_one.tolist())],
                                out_array[tuple(one_to_end.tolist())]]), 0)

    return out_array


def CenterTime(a):
    return interior_vertex_func(a, dims=(0,), func=mean)


def CenterLay(a):
    return interior_vertex_func(a, dims=(1,), func=mean)


def CenterRow(a):
    return interior_vertex_func(a, dims=(2,), func=mean)


def CenterCol(a):
    return interior_vertex_func(a, dims=(3,), func=mean)


def CenterRowCol(a):
    return interior_vertex_func(a, dims=(2, 3), func=mean)


def CenterTimeRowCol(a):
    return interior_vertex_func(a, dims=(0, 2, 3), func=mean)


CenterCMAQWind = CenterTimeRowCol


def CenterCAMxWind(a):
    out = zeros(a.shape, 'f')
    out[:, :, 1:-1, 1:-1] = CenterRowCol(a[:, :, :-1, :-1])
    out[:, :, [0, -1], 1:-1] = CenterCol(a[:, :, [0, -1], :-1])
    out[:, :, 1:-1, [0, -1]] = CenterRow(a[:, :, :-1, [0, -1]])
    out[:, :, [0, -1], [0, -1]] = a[:, :, [0, -2], [0, -2]]
    return CenterTime(out)


def CenterCAMxU(a):
    out = zeros(a.shape, 'f')
    out[:, :, :, 1:-1] = CenterCol(a[:, :, :, :-1])
    return CenterTime(out)


def CenterCAMxV(a):
    out = zeros(a.shape, 'f')
    out[:, :, 1:-1, :] = CenterRow(a[:, :, :-1, :])
    return CenterTime(out)


def BoundToDiff(a, dim=-1):
    top = [slice(None)] * len(a.shape)
    top[dim] = slice(1, None)

    bottom = [slice(None)] * len(a.shape)
    bottom[dim] = slice(None, -1)
    return a[top] - a[bottom]


def CAMxHeightToDepth(a):
    b = a.copy()
    b[:, 1:, :, :] = BoundToDiff(a, 1)
    return b


def ConvertCAMxTime(date, time, nvars):
    class temp:
        pass

    f = temp()
    f.dimensions = {'TSTEP': date.shape[0], 'VAR': nvars, 'DATE-TIME': 2}

    a = array([date, time], dtype='i').swapaxes(0, 1)
    if len(a.shape) == 2:
        a = a[:, newaxis, :]
    date = a[:, :, 0]
    if (date < 70000).any():
        date += 2000000
    else:
        date += 1900000
    time = a[:, :, 1]
    while not (time == 0).all() and time.max() < 10000:
        time *= 100
    a = PseudoNetCDFVariable(
        f, 'TFLAG', 'i', ('TSTEP', 'VAR', 'DATE-TIME'),
        values=a[:, [0], :].repeat(nvars, 1))
    a.units = 'DATE-TIME'.ljust(16)
    a.long_name = 'TFLAG'.ljust(16)
    a.var_desc = a.long_name
    return a


class TestInteriorVertex(unittest.TestCase):
    def setUp(self):
        self.A2D = arange(1, 17).reshape(4, 4)
        self.A3D = arange(1, 17).reshape(1, 4, 4).repeat(4, 0)
        self.A4D = arange(1, 17).reshape(1, 1, 4, 4).repeat(4, 0).repeat(4, 1)
        self.mean2D = array([[3.5, 4.5, 5.5],
                             [7.5, 8.5, 9.5],
                             [11.5, 12.5, 13.5]])

        self.sum2D = array([[14, 18, 22],
                            [30, 34, 38],
                            [46, 50, 54]])

    def runTest(self):
        pass

    def test2D2DSum(self):
        self.assertTrue((interior_vertex_func(
            self.A2D, dims=(-1, -2), func=sum) == self.sum2D).all())

    def test2D2DMean(self):
        self.assertTrue((interior_vertex_func(
            self.A2D, dims=(-1, -2), func=mean) == self.mean2D).all())

    def test3D2DSum(self):
        newv = interior_vertex_func(self.A3D, dims=(-1, -2), func=sum)
        checkv = self.sum2D.reshape(1, 3, 3)
        self.assertTrue((newv == checkv).all())

    def test3D2DMean(self):
        newv = interior_vertex_func(self.A3D, dims=(-1, -2), func=mean)
        checkv = self.mean2D.reshape(1, 3, 3)
        self.assertTrue((newv == checkv).all())

    def test3D3DSum(self):
        newv = interior_vertex_func(self.A3D, dims=(-1, -2, -3), func=sum)
        checkv = (self.sum2D.reshape(1, 3, 3) * 2)
        self.assertTrue((newv == checkv).all())

    def test3D3DMean(self):
        newv = interior_vertex_func(self.A3D, dims=(-1, -2, -3), func=mean)
        checkv = self.mean2D.reshape(1, 3, 3)
        self.assertTrue((newv == checkv).all())

    def test4D2DSum(self):
        newv = interior_vertex_func(self.A4D, dims=(-1, -2), func=sum)
        checkv = self.sum2D.reshape(1, 1, 3, 3)
        self.assertTrue((newv == checkv).all())

    def test4D2DMean(self):
        newv = interior_vertex_func(self.A4D, dims=(-1, -2), func=mean)
        checkv = self.mean2D.reshape(1, 1, 3, 3)
        self.assertTrue((newv == checkv).all())

    def test4D3DSum(self):
        newv = interior_vertex_func(self.A4D, dims=(-1, -2, -3), func=sum)
        checkv = self.sum2D.reshape(1, 1, 3, 3) * 2
        self.assertTrue((newv == checkv).all())

    def test4D3DMean(self):
        newv = interior_vertex_func(self.A4D, dims=(-1, -2, -3), func=mean)
        checkv = self.mean2D.reshape(1, 1, 3, 3)
        self.assertTrue((newv == checkv).all())

    def test4D4DSum(self):
        newv = interior_vertex_func(self.A4D, dims=(-1, -2, -3, -4), func=sum)
        checkv = (self.sum2D.reshape(1, 1, 3, 3) * 4)
        self.assertTrue((newv == checkv).all())

    def test4D4DMean(self):
        newv = interior_vertex_func(self.A4D, dims=(-1, -2, -3, -4), func=mean)
        checkv = self.mean2D.reshape(1, 1, 3, 3)
        self.assertTrue((newv == checkv).all())


if __name__ == '__main__':
    unittest.main()



import numpy as np
from PseudoNetCDF._getwriter import registerwriter

"""
hour,idate,((temps(i,j),i=1,nx),j=1,ny)
Loop from 1 to nlay layers:
    hour,idate,((temp(i,j,k),i=1,nx),j=1,ny)
"""


def ncf2temperature(ncffile, outpath):
    outfile = open(outpath, 'wb')
    sfc = ncffile.variables['SURFTEMP']
    air = ncffile.variables['AIRTEMP']
    nz, nr, nc = air.shape[-3:]
    nelem = nr * nc * 4 + 8
    for di, (d, t) in enumerate(ncffile.variables['TFLAG'][:, 0]):
        t = np.array(t / 100, ndmin=1, dtype='>f')
        d = np.array(d, ndmin=1).astype('>i')
        d = (d % (d // 100000 * 100000)).astype('>i')
        buf = np.array(nelem, dtype='>i').tostring()
        outfile.write(buf)
        t.tofile(outfile)
        d.tofile(outfile)
        sfc[di].astype('>f').tofile(outfile)
        outfile.write(buf)
        for zi in range(nz):
            outfile.write(buf)
            t.tofile(outfile)
            d.tofile(outfile)
            air[di, zi].astype('>f').tofile(outfile)
            outfile.write(buf)

    outfile.flush()
    return outfile


registerwriter('camxfiles.temperature', ncf2temperature)
registerwriter('temperature', ncf2temperature)



import textwrap


def griddes(infile):
    gridparams = {}
    gridparams['GTYPE'] = 'unstructured'
    gridparams['NX'] = len(infile.dimensions['longitude'])
    gridparams['NY'] = len(infile.dimensions['latitude'])
    gridparams['XUNIT'] = 'degrees_east'
    gridparams['YUNIT'] = 'degrees_north'
    exec('NCELLS = NX * NY', None, gridparams)
    LONSTR = textwrap.wrap(
        ' '.join(['%f' % lon for lon in infile.variables['longitude'][:]]), 80)
    LATSTR = textwrap.wrap(
        ' '.join(['%f' % lat for lat in infile.variables['latitude'][:]]), 80)
    gridparams['LONSTR'] = LONSTR
    gridparams['LATSTR'] = LATSTR
    out = """
    gridtype  = %(GTYPE)s
    gridsize  = %(NCELLS)d
    xsize     = %(NX)d
    ysize     = %(NY)d
    xunits    = %(XUNIT)s
    yunits    = %(YUNIT)s
    xvals     = %(LONSTR)s
    yvals     = %(LATSTR)s
    """ % gridparams
    return out



__all__ = ['test_bpch', 'test_bpch1', 'test_bpch2', 'test_geos']


from PseudoNetCDF.geoschemfiles._bpch import TestMemmaps as test_bpch1
from PseudoNetCDF.geoschemfiles._newbpch import TestMemmaps as test_bpch2
from PseudoNetCDF.geoschemfiles._bpchmaster import TestMemmaps as test_bpch
from PseudoNetCDF.geoschemfiles._geos import TestMemmaps as test_geos



import numpy as np
from datetime import datetime, timedelta
import re


def get_datetime_from_relative(var, format='%Y-%m-%d %H:%M:%S'):
    units = var.units
    tunit, refstr = units.split(' since ')
    refstr = re.sub(r'\s*(UTC)?', '', refstr)
    rdate = datetime.strptime(refstr, format)

    out = np.array([rdate + timedelta(**{tunit: float(dt)}) for dt in var[:]])
    return out



__all__ = ['wind']
__doc__ = """
.. _Memmap
:mod:`Memmap` -- wind Memmap interface
============================================

.. module:: Memmap
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` memory map for CAMx
              wind files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

# Distribution packages
import unittest
import struct

# Site-Packages
from numpy import zeros, array, memmap, nan

# This Package modules
from PseudoNetCDF.camxfiles.FortranFileUtil import OpenRecordFile
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariable
from PseudoNetCDF.sci_var import PseudoNetCDFVariables
from collections import OrderedDict
from PseudoNetCDF.ArrayTransforms import ConvertCAMxTime

# for use in identifying uncaught nan
listnan = struct.unpack('>f', b'\xff\xc0\x00\x00')[0]
checkarray = zeros((1,), 'f')
checkarray[0] = listnan
array_nan = checkarray[0]


class wind(PseudoNetCDFFile):
    """
    wind provides a PseudoNetCDF interface for CAMx
    wind files.  Where possible, the inteface follows
    IOAPI conventions (see www.baronams.com).

    ex:
        >>> wind_path = 'camx_wind.bin'
        >>> rows,cols = 65,83
        >>> windfile = wind(wind_path,rows,cols)
        >>> windfile.variables.keys()
        ['TFLAG', 'U', 'V']
        >>> v = windfile.variables['V']
        >>> tflag = windfile.variables['TFLAG']
        >>> tflag.dimensions
        ('TSTEP', 'VAR', 'DATE-TIME')
        >>> tflag[0,0,:]
        array([2005185,       0])
        >>> tflag[-1,0,:]
        array([2005185,  240000])
        >>> v.dimensions
        ('TSTEP', 'LAY', 'ROW', 'COL')
        >>> v.shape
        (25, 28, 65, 83)
        >>> windfile.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 65, 'COL': 83}
    """

    __units = 'm/s'

    def __init__(self, rffile, rows, cols):
        rf = OpenRecordFile(rffile)
        self.__time_hdr_fmts = {12: "fii", 8: "fi"}[rf.record_size]
        self.__time_hdr_fmts_size = rf.record_size
        self.STIME, self.SDATE = rf.unpack("fi")

        rf.next()
        lays = 1
        record_size = rf.record_size
        while rf.record_size == record_size:
            lays += 1
            rf.next()
        self.__dummy_length = (rf.record_size + 8) // 4
        lays //= 2
        record = rows * cols * 4 + 8
        total_size = self.__dummy_length
        times = 0
        while total_size < rf.length:
            times += 1
            total_size += record * 2 * lays + self.__time_hdr_fmts_size + 8
        times -= 1

        self.variables = OrderedDict
        del rf
        self.createDimension('TSTEP', times)
        self.createDimension('DATE-TIME', 2)
        self.createDimension('LAY', lays)
        self.createDimension('ROW', rows)
        self.createDimension('COL', cols)
        self.createDimension('VAR', 2)

        self.NVARS = len(self.dimensions['VAR'])
        self.NLAYS = len(self.dimensions['LAY'])
        self.NROWS = len(self.dimensions['ROW'])
        self.NCOLS = len(self.dimensions['COL'])
        self.FTYPE = 1

        self.__memmap = memmap(rffile, '>f', 'r', offset=0)

        if self.__time_hdr_fmts_size == 12:
            self.LSTAGGER = self.__memmap[3].view('i')
        else:
            self.LSTAGGER = nan

        self.variables = PseudoNetCDFVariables(
            self.__variables, ['TFLAG', 'U', 'V'])

    def __variables(self, k):
        self.__add_variables()
        return self.variables[k]

    def __decorator(self, k, pncfv):
        decor = {'TFLAG': {'units': 'DATE-TIME',
                           'long_name': 'TFLAG', 'var_desc': 'Time flag'}}
        for k, v in decor.get(k, {'units': 'm/s', 'long_name': k,
                                  'var_desc': k}).items():
            setattr(pncfv, k, v)
        return pncfv

    def __add_variables(self):
        tsteps = len(self.dimensions['TSTEP'])
        lays = len(self.dimensions['LAY'])
        rows = len(self.dimensions['ROW'])
        cols = len(self.dimensions['COL'])
        offset = len(self.__time_hdr_fmts) + 2
        block = (rows * cols + 2) * 2 * lays
        out_idx = zeros(self.__memmap.shape, 'b')
        for t in range(tsteps):
            start = (t + 1) * offset + t * block + t * self.__dummy_length
            stop = start + block

            nlay2 = lays * 2
            nmatrix = rows * cols
            out_idx[start:stop].reshape(nlay2, nmatrix + 2)[:, 1:-1] = 1
            out_idx[start:stop].reshape(nlay2, nmatrix + 2)[:, [0, -1]] = 2
            out_idx[start - offset:start] = 3

        buffer = self.__memmap[out_idx == 2].reshape(tsteps, lays, 2, 2)
        if not (buffer[:, :, :, 0] == buffer[:, :, :, 1]).all():
            raise ValueError('Fortran unformatted record start and end ' +
                             'padding do not match.')
        date = self.__memmap[out_idx == 3].reshape(
            tsteps, (out_idx == 3).sum() // tsteps)[:, 2].view('>i')
        time = self.__memmap[out_idx == 3].reshape(
            tsteps, (out_idx == 3).sum() // tsteps)[:, 1]

        self.variables['TFLAG'] = ConvertCAMxTime(date, time, 2)
        uvals = self.__memmap[out_idx == 1].reshape(tsteps, lays, 2, rows,
                                                    cols)[:, :, 0, :, :]
        self.variables['U'] = self.__decorator('U', PseudoNetCDFVariable(
            self, 'U', 'f', ('TSTEP', 'LAY', 'ROW', 'COL'),
            values=uvals))
        vvals = self.__memmap[out_idx == 1].reshape(tsteps, lays, 2, rows,
                                                    cols)[:, :, 1, :, :]
        self.variables['V'] = self.__decorator('V', PseudoNetCDFVariable(
            self, 'V', 'f', ('TSTEP', 'LAY', 'ROW', 'COL'), values=vvals))


class TestMemmap(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testWD(self):
        import PseudoNetCDF.testcase
        wdfile = wind(PseudoNetCDF.testcase.camxfiles_paths['wind'], 4, 5)
        wdfile.variables['TFLAG']
        checkv = array([
            -1.73236704e+00, -1.99612117e+00, -3.00912833e+00, -3.92667437e+00,
            -3.49521232e+00, 2.04542422e+00, 8.57666790e-01, -1.71201074e+00,
            -4.24386787e+00, -5.37704515e+00, 1.85697508e+00, 6.34313405e-01,
            -1.21529281e+00, -3.03180861e+00, -4.36278439e+00, -1.90753967e-01,
            -1.08261776e+00, -1.73634803e+00, -2.10829663e+00, -2.28424144e+00,
            -1.88443780e+00, -2.02582169e+00, -3.09955978e+00, -4.14587784e+00,
            -3.72402787e+00, 2.16277528e+00, 8.94082963e-01, -1.86343944e+00,
            -4.58147812e+00, -5.81837606e+00, 1.97949493e+00, 6.12511635e-01,
            -1.35096896e+00, -3.25313163e+00, -4.67790413e+00, -1.89851984e-01,
            -1.16381800e+00, -1.84269297e+00, -2.21348834e+00, -2.40952253e+00,
            -2.04972148e+00, -2.11795568e+00, -3.06094027e+00, -4.11207581e+00,
            -3.74964952e+00, 2.09780049e+00, 8.01259458e-01, -1.90404522e+00,
            -4.59170580e+00, -5.83114100e+00, 1.97475648e+00, 5.54396451e-01,
            -1.41695607e+00, -3.28227353e+00, -4.67724609e+00, -1.94723800e-01,
            -1.18353117e+00, -1.86556363e+00, -2.22842574e+00, -2.42080784e+00,
            -1.65720737e+00, -1.58054411e+00, -2.25336742e+00, -3.06462526e+00,
            -2.47261453e+00, 1.37642264e+00, 1.16142654e+00, -6.82058990e-01,
            -2.68112469e+00, -3.38680530e+00, 1.80796599e+00, 1.48641026e+00,
            -1.71508826e-02, -1.68607295e+00, -2.89399385e+00, 3.40398103e-01,
            3.25049832e-02, -5.91206312e-01, -1.19038010e+00, -1.52301860e+00,
            -1.83006203e+00, -1.74505961e+00, -2.50190806e+00, -3.29507184e+00,
            -2.65367699e+00, 1.55719578e+00, 1.25234461e+00, -9.14537191e-01,
            -3.16307521e+00, -4.00584650e+00, 2.07018161e+00, 1.60957754e+00,
            -1.46312386e-01, -2.04018188e+00, -3.40377665e+00, 4.11731720e-01,
            -2.29119677e-02, -7.27373540e-01, -1.35116744e+00, -1.70711970e+00,
            -1.72859466e+00, -1.73683071e+00, -2.65253377e+00, -3.43689489e+00,
            -2.75470304e+00, 1.58366191e+00, 1.19656324e+00, -1.09935236e+00,
            -3.38544369e+00, -4.26436615e+00, 2.04826832e+00, 1.53576791e+00,
            -2.60809243e-01, -2.18679833e+00, -3.59082842e+00, 3.77060443e-01,
            -1.05680525e-01, -8.10511589e-01, -1.40993130e+00,
            -1.76300752e+00], dtype='f').reshape(2, 3, 4, 5)
        self.assertTrue((wdfile.variables['V'][:] == checkv).all())

    def testNCF2WD(self):
        import PseudoNetCDF.testcase
        from PseudoNetCDF.pncgen import pncgen

        import os
        inpath = PseudoNetCDF.testcase.camxfiles_paths['wind']
        outpath = PseudoNetCDF.testcase.camxfiles_paths['wind'] + '.check'
        infile = wind(inpath, 4, 5)
        pncgen(infile, outpath, format='camxfiles.wind')
        orig = open(inpath, 'rb').read()
        new = open(outpath, 'rb').read()
        os.remove(outpath)
        assert(orig == new)


TestSuite = unittest.makeSuite(TestMemmap, 'test')
if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python


def main():
    from PseudoNetCDF.pncparse import getparser, pncparse
    from PseudoNetCDF.plotutil.pncts import plotts
    parser = getparser(plot_options=True, has_ofile=True)
    parser.epilog += """
    -----
box.py inobs inmod target [target ...]
inobs - path to obs
inmod - path to mod
target - variable name
"""
    ifiles, args = pncparse(plot_options=True, has_ofile=True, parser=parser)
    plotts(args)


if __name__ == '__main__':
    main()



from warnings import warn
from collections import OrderedDict

import numpy as np


class PseudoNetCDFVariable(np.ndarray):
    """
    PseudoNetCDFVariable presents the Scientific.IO.NetCDF.NetCDFVariable
    interface, but unlike that type, provides a contructor for variables
    that could be used without adding it to the parent file
    """
    __array_priority__ = 10000000.

    def xarray(self, iscoord=False):
        """
        Experimental function

        Returns
        -------
        out : xarray.DataArray
            object with dimensions and coordinates
            dims : string are set by self.get_coords()
            coords : are set by self.get_coords()

        Notes
        -----
        When data is 2-d, the dimensions
        """
        import xarray as xr
        myname = self._name
        if iscoord:
            coordd = {}
        else:
            coordd = dict([(k, self.get_coord(k))
                           for k in self.get_coord_names() if k != myname])
        attrs = dict([(k, getattr(self, k)) for k in self.ncattrs()])
        return xr.DataArray(self[:], dims=self.dimensions, coords=coordd,
                            attrs=attrs)

    def get_coord_names(self):
        dims = list(self.dimensions)
        ndims = len(dims)
        coords = getattr(self, 'coordinates', '').split()
        ncoords = len(coords)
        maxdims = max(ndims, ncoords)
        coords = ((ndims * [None]) + coords)[-maxdims:]
        dims = (dims + ncoords * [None])[:maxdims]
        coords = [dimn if coordn is None else coordn for dimn,
                  coordn in zip(dims, coords)]
        return coords

    def get_coord(self, coordn):
        if coordn in self._parent.variables:
            v = self._parent.variables[coordn]
            if len(v.dimensions) > 1 and coordn in v.dimensions:
                warn(coordn + ' has dimensions ' + str(v.dimensions) +
                     '; so it will use a standard linear coordinate. ' +
                     'Consider manually adding coords')
                coordi = list(self.dimensions).index(coordn)
                coordv = np.arange(self.shape[coordi])
            else:
                coordv = (v.dimensions, v.xarray(iscoord=True))
        else:
            coordi = list(self.dimensions).index(coordn)
            coordv = np.arange(self.shape[coordi])
        return coordv

    def get_coords(self):
        coords = self.get_coord_names()
        coordsd = {}
        for coordn in coords:
            coordsd[coordn] = self.get_coord(coordn)
        return coordsd

    def __repr__(self):
        out = str(self)
        return object.__repr__(self).replace(' at ', '\n' + out + ' at ')

    def __str__(self):
        namekeys = ['_name', 'name', 'standard_name', 'long_name']
        if hasattr(self, '_name'):
            var_name = getattr(self, '_name').strip()
        else:
            for nck in self.ncattrs():
                if 'name' in nck:
                    namekeys.append(nck)

            for nck in namekeys:
                if hasattr(self, nck):
                    var_name = getattr(self, nck).strip()
                    break
            else:
                var_name = 'unknown'

        var_type = dict(float32='float',
                        float64='double',
                        int32='integer',
                        uint32='integer',
                        int64='long',
                        bool='bool',
                        string8='char',
                        string80='char').get(self.dtype.name, self.dtype.name)
        out = ""
        indent = "    "
        startindent = ""
        out += (
            startindent + 1 * indent +
            ("%s %s%s; // shape: %s\n" % (var_type, var_name,
                                          str(self.dimensions)
                                          .replace('u\'', '')
                                          .replace('\'', '')
                                          .replace(',)', ')'),
                                          str(self.shape))))
        for prop_name in self.ncattrs():
            prop = getattr(self, prop_name)
            out += startindent + 2 * indent + \
                ("%s:%s = %s ;\n" %
                 (var_name, prop_name, repr(prop).replace("'", '"')))

        out += 'array: '
        out += self.array().__str__()
        return out

    def array(self):
        """
        Returns
        -------
        out : numpy.ndarray
            parent type view object
        """
        return self.view(type=np.ndarray)

    def __setattr__(self, k, v):
        """
        Set attributes (aka properties) and identify user-defined attributes.
        """
        if k[:1] != '_' and \
           k not in ('dimensions', 'typecode'):
            if k not in self._ncattrs:
                self._ncattrs += (k, )
        object.__setattr__(self, k, v)

    def __delattr__(self, k):
        if k in self._ncattrs:
            self._ncattrs = tuple([k_ for k_ in self._ncattrs if k_ != k])
        object.__delattr__(self, k)

    def setncattr(self, k, v):
        return setattr(self, k, v)

    def getncattr(self, k):
        return getattr(self, k)

    def setncatts(self, attdict):
        for pk, pv in attdict.items():
            setattr(self, pk, pv)

    def getncatts(self):
        outd = OrderedDict()
        for pk in self.ncattrs():
            outd[pk] = getattr(self, pk)
        return outd

    def ncattrs(self):
        """
        Returns
        -------
        ncattrs : tuple of strings
            attributes that have been user defined
        """

        return self._ncattrs

    def __new__(subtype, parent, name, typecode, dimensions, **kwds):
        """
        Creates a variable using the dimensions as defined in
        the parent object

        Parameters
        ----------
        parent : PseudoNetCDFFile
            an object with a dimensions variable
        name: string
            name for variable
        typecode: string or numpy.dtype
            numpy style typecode
        dimensions: tuple of strings
            dimension names to be used from parent
        **kwds: keywords
            Dictionary of keywords to be added as properties to the variable.
            The keyword 'values' is a special case that will be used as the
            starting values of the array

        """
        if 'values' in kwds.keys():
            result = kwds.pop('values')
            typecode = result.dtype.char
            # Adding easy default dimension object;
            # Avoiding PseudoNetCDFFile due to recursive dependence
            if parent is None:
                from ._files import PseudoNetCDFFile
                parent = PseudoNetCDFFile()
                for dk, dl in zip(dimensions, result.shape):
                    parent.createDimension(dk, dl)
        else:
            shape = []
            for d in dimensions:
                dim = parent.dimensions[d]

                # Adding support for netCDF3 dimension objects
                if not isinstance(dim, int):
                    dim = len(dim)
                shape.append(dim)

            result = np.zeros(shape, typecode)

        result = result[...].view(subtype)

        result.typecode = lambda: typecode
        result.dimensions = tuple(dimensions)
        result._ncattrs = ()
        result._parent = parent
        result._name = name
        for k, v in kwds.items():
            setattr(result, k, v)
        return result

    def __array_finalize__(self, obj):
        if obj is None:
            return
        _parent = getattr(obj, '_parent', object())
        object.__setattr__(self, '_parent', _parent)
        _name = getattr(obj, '_name', 'unknown')
        object.__setattr__(self, '_name', _name)
        ntypecode = getattr(obj, 'typecode', lambda: self.dtype.char)
        object.__setattr__(self, 'typecode', ntypecode)
        ndimensions = getattr(obj, 'dimensions', lambda: self.dtype.char)
        object.__setattr__(self, 'dimensions', ndimensions)
        nncattrs = getattr(obj, '_ncattrs', getattr(self, '_ncattrs', ()))
        object.__setattr__(self, '_ncattrs', nncattrs)
        if hasattr(obj, '_ncattrs'):
            for k in nncattrs:
                if not hasattr(self, k):
                    object.__setattr__(self, k, getattr(obj, k))

    def swapaxes(self, a1, a2):
        out = np.ndarray.swapaxes(self, a1, a2)
        newdims = list(self.dimensions)
        newdims[a1] = self.dimensions[a2]
        newdims[a2] = self.dimensions[a1]
        out.dimensions = tuple(newdims)
        return out

    def getValue(self):
        """
        Returns
        -------
        item : scalar
            value
        """
        return self.item()

    def assignValue(self, value):
        """
        Parameters
        ----------
        value : scalar
            assign value to scalar variable
        """
        self.itemset(value)


class PseudoNetCDFMaskedVariable(PseudoNetCDFVariable, np.ma.MaskedArray):
    __array_priority__ = 1000000000.

    def __new__(subtype, parent, name, typecode='f', dimensions=(), **kwds):
        """
        Creates a variable using the dimensions as defined in
        the parent object

        parent: PseudoNetCDFFile
            an object with a dimensions variable
        name: string
            name for variable
        typecode: string or numpy.dtype
            numpy style typecode
        dimensions: tuple of strings
            dimension names to be used from parent
        **kwds: dictionary
            keywords to be added as properties to the variable. The keyword
            'values' is a special case that will be used as the starting
            values of the array

        """
        if 'values' in kwds.keys():
            result = kwds.pop('values')
        else:
            shape = []
            for d in dimensions:
                dim = parent.dimensions[d]

                # Adding support for netCDF3 dimension objects
                if not isinstance(dim, int):
                    dim = len(dim)
                shape.append(dim)
            fill_value = kwds.get('fill_value', None)
            if fill_value is None:
                for pk in ('fill_value', 'missing_value', '_FillValue'):
                    fill_value = kwds.get(pk, None)
                    if fill_value is not None:
                        break
            result = np.ma.zeros(shape, dtype='S1' if typecode ==
                                 'c' else typecode,
                                 fill_value=fill_value)

        result = result.view(subtype)
        result._ncattrs = ()
        result._parent = parent
        result._name = name
        result.typecode = lambda: typecode
        result.dimensions = tuple(dimensions)
        for k, v in kwds.items():
            setattr(result, k, v)
        return result

    def array(self):
        """
        Returns
        -------
        out : numpy.ma.masked_array
            parent type view object
        """
        return self.view(type=np.ma.masked_array)

    def __array_finalize__(self, obj):
        np.ma.MaskedArray.__array_finalize__(self, obj)

    def _update_from(self, obj):
        dt = self.dtype.char
        self.typecode = getattr(
            obj, 'typecode', lambda: ('c' if dt == 'S' else dt))
        self.dimensions = getattr(
            obj, 'dimensions', getattr(self, 'dimensions', ()))
        self._ncattrs = getattr(obj, '_ncattrs', getattr(self, '_ncattrs', ()))
        self._fill_value = getattr(
            obj, '_fill_value', getattr(self, '_fill_value', -999))
        if hasattr(obj, '_ncattrs'):
            for k in obj._ncattrs:
                if k in ('fill_value',):
                    continue
                setattr(self, k, getattr(obj, k))
        np.ma.MaskedArray._update_from(self, obj)

    def __getitem__(self, item):
        out = np.ma.MaskedArray.__getitem__(self, item)
        try:
            out._fill_value = self._fill_value
        except Exception:
            pass
        out = out.view(PseudoNetCDFMaskedVariable)
        if np.isscalar(out):
            return out
        if hasattr(self, 'dimensions'):
            out.dimensions = self.dimensions
        if hasattr(self, '_ncattrs'):
            for k in self._ncattrs:
                setattr(out, k, getattr(self, k))
        else:
            self._ncattrs = ()
        return out

    def __setattr__(self, k, v):
        """
        Set attributes (aka properties) and identify user-defined attributes.
        """
        if k[:1] != '_' and \
           k not in ('dimensions', 'typecode', 'mask'):
            if k not in self._ncattrs:
                self._ncattrs += (k, )
        np.ma.MaskedArray.__setattr__(self, k, v)

    def __delattr__(self, k):
        if k in self._ncattrs:
            self._ncattrs = tuple([k_ for k_ in self._ncattrs if k_ != k])
        object.__delattr__(self, k)

    def swapaxes(self, a1, a2):
        out = np.ma.masked_array.swapaxes(self, a1, a2)
        newdims = list(self.dimensions)
        newdims[a1] = self.dimensions[a2]
        newdims[a2] = self.dimensions[a1]
        out.dimensions = tuple(newdims)
        return out

    def ncattrs(self):
        """
        Returns
        -------
        ncattrs : tuple of strings
            attributes that have been user defined
        """

        return self._ncattrs

    def getValue(self):
        """
        Returns
        -------
        out : scalar
            value
        """
        return self.item()

    def assignValue(self, value):
        """
        Parameters
        ----------
        value : scalar
            value to assign to scalar variable
        """
        self.itemset(value)


def PseudoIOAPIVariable(parent, name, typecode, dimensions, **kwds):
    """
    Creates a variable using the dimensions as defined in
    the parent object

    Parameters
    ----------
    parent : PseudoNetCDFFile
        an object with a dimensions variable
    name: string
        name for variable
    typecode: string or numpy.dtype
        numpy style typecode
    dimensions: tuple of strings
        dimension names to be used from parent
    **kwds : dictionary
        units: default = none
        long_name: default = name
        var_desc: default = name
    """

    retval = PseudoNetCDFVariable(parent, name, typecode, dimensions, **kwds)

    if 'units' not in kwds:
        warn('IOAPI variables must have units; %s has been initialized ' +
             'with "None" units')
        retval.units = 'None'

    if 'long_name' not in kwds:
        retval.long_name = name.ljust(16)

    if 'var_desc' not in kwds:
        retval.var_desc = name.ljust(80)

    return retval



from numpy import memmap, dtype
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariables
from PseudoNetCDF.sci_var import PseudoNetCDFVariable
import os


class greasd(PseudoNetCDFFile):
    def __init__(self, pig_path, mode='r'):
        file_size = os.path.getsize(pig_path)
        tmpmap = memmap(pig_path, dtype='>i', shape=7)
        self.SDATE = tmpmap[1]
        self.STIME = tmpmap[2:3].view('>f')[0]
        self.NPIG = tmpmap[3]
        record_length = int(tmpmap[-1] / 4)
        per_pig = record_length / self.NPIG

        pig_dtype = dtype(dict(
            names='SPAD1 IDATE TIME NPIG IDUM EPAD1'.split() +
                  ['SPAD2'] +
                  ['P%d_%d' % (i / per_pig, i % per_pig)
                   for i in range(record_length)] +
                  ['EPAD2'],
            formats='>i >i >f >i >i >i >i'.split() +
                    ['>f'] * record_length + ['>i']))
        items = float(file_size) / float(pig_dtype.itemsize)
        assert(items == int(items))
        self.NSTEPS = items = int(items)
        self.__memmap__ = memmap(pig_path, shape=items, dtype=pig_dtype)
        self.dimensions = dict(TSTEP=items, NPIG=self.NPIG)
        self.variables = PseudoNetCDFVariables(
            self.__variables, pig_dtype.fields.keys())

    def __variables(self, k):
        return PseudoNetCDFVariable(self, k, 'f', ('TSTEP',),
                                    values=self.__memmap__[k],
                                    var_desc=k.ljust(16),
                                    long_name=k.ljust(16))



__doc__ = """
.. _Readers
:mod:`Readers` -- CAMx Reader Interfaces
========================================

.. module:: Readers
   :platform: Unix, Windows
   :synopsis: Provides an easy access point for disk read based
              :ref:`PseudoNetCDF` file interfaces for all CAMx
              related files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
__all__ = ['height_pressure',
           'humidity',
           'ipr',
           'irr',
           'one3d',
           'point_source',
           'temperature',
           'uamiv',
           'vertical_diffusivity',
           'wind']

from .height_pressure.Read import height_pressure
from .humidity.Read import humidity
from .ipr.Read import ipr
from .irr.Read import irr
from .one3d.Read import one3d
from .point_source.Read import point_source
from .temperature.Read import temperature
from .uamiv.Read import uamiv
from .vertical_diffusivity.Read import vertical_diffusivity
from .wind.Read import wind



from PseudoNetCDF.sci_var import PseudoNetCDFMaskedVariable
import numpy as np

PseudoNetCDFVariable = PseudoNetCDFMaskedVariable


def add_derived_met(ifile, func, *args, **kwds):
    results = func(*args, **kwds)
    if not isinstance(results, tuple):
        results = (results,)
    for result in results:
        ifile.variables[result.short_name] = result


def wmr_ptd(p, td, gkg=False):
    """
    Inputs:
       p = surface pressure in mb;
       Td = dew point in deg C;
    Calculates
       e = vapor pressure in mb;

    Returns
       q = specific humidity in kg/kg.
    """
    dimensions = getattr(p, 'dimensions', getattr(td, 'dimensions', 'unknown'))
    e = 6.112 * np.ma.exp((17.67 * td) / (td + 243.5))
    q = (0.622 * e) / (p - (0.378 * e))
    if gkg:
        q *= 1000.
        units = 'g/kg'
    else:
        units = 'kg/kg'
    Q = PseudoNetCDFVariable(None, 'Q', q.dtype.char, dimensions, units=units,
                             long_name='specific humidity in kg/kg',
                             short_name='Q', values=q)
    return Q


def relhum_ttd(t, td, percent=False, add=True):
    """
    t - temperature in K
    td - dewpoint temperature in K


    ;
    ; Calculate relative humidity given temperature (K)
    ; and dew point temperature (K)
    ;
    ; reference: John Dutton, Ceaseless Wind, 1976
    """
    dimensions = getattr(t, 'dimensions', getattr(
        td, 'dimensions', ('unknown',)))
    gc = 461.5                        # [j/{kg-k}]   gas constant water vapor
    gc = gc / (1000. * 4.186)             # [cal/{g-k}]  change units
    # lhv=latent heat vap
    lhv = (597.3 - 0.57 * (t - 273.))      # dutton top p273 [empirical]

    values = np.ma.exp((lhv / gc) * (1.0 / t - 1.0 / td))
    if percent:
        values *= 100.

    rh = PseudoNetCDFVariable(None, 'RH', 'f', dimensions, values=values)
    rh.long_name = "relative humidity"
    if percent:
        rh.units = "%"
    else:
        rh.units = "fraction"
    rh.short_name = 'RH'
    return rh


def wmrq_ptd(p, td, dry=False, kgkg=False):
    """
    p - pressure in Pa
    td - temperature in degrees celcius
    dry - return results in mixing ratio or specific humidity in dry air
    kgkg - return results in kg/kg instead of g/kg
    """
    dimensions = getattr(p, 'dimensions', getattr(
        td, 'dimensions', ('unknown',)))
    p = np.asarray(p)
    td = np.asarray(td)
    # ncl:   q = mixhum_ptd (p,td,option)   [ q=g/kg ]
    #                        p  - PA
    #                        td - K
    # local
    # INTEGER  N
    # DOUBLE PRECISION T0, PA2MB
    # DATA     T0    /273.15d0/
    # DATA     PA2MB /0.01d0  /
    T0 = 273.15
    PA2MB = .01

    # mixing ratio (kg/kg)
    # the function wants hPA (mb) and degrees centigrade
    WMR = wmr_skewt_pt(p * PA2MB, td - T0) * 0.001

    # if iswit=2 calculate specific humidity (kg/kg)

    if dry:
        name = 'W'
        long_name = 'mass of water per mass of dry air'
    else:
        name = 'Q'
        long_name = 'mass of water per mass of air (dry+wet)'
        WMR = WMR / (WMR + 1.0)

    # if iswit < 0 then return g/kg

    if kgkg:
        WMR *= 1000.
        units = 'g/kg'
    else:
        units = 'kg/kg'
    wmr = PseudoNetCDFVariable(None, name, 'f', dimensions, values=WMR,
                               units=units, long_name=long_name)
    wmr.short_name = name
    return wmr


def wmr_skewt_pt(p, t):
    """
    p - pressure in mb
    t - temperature in degrees C
    this function approximates the mixing ratio wmr (grams of water
    vapor per kilogram of dry air) given the pressure p (mb) and the
    temperature t (celsius). the formula used is given on p. 302 of the
    smithsonian meteorological tables by roland list (6th edition).
    """
    # eps = ratio of the mean molecular weight of water (18.016 g/mole)
    #      to that of dry air (28.966 g/mole)
    EPS = 0.62197

    # the next two lines contain a formula by herman wobus for the
    # correction factor wfw for the departure of the mixture of air
    # and water vapor from the ideal gas law. the formula fits values
    # in table 89, p. 340 of the smithsonian meteorological tables,
    # but only for temperatures and pressures normally encountered in
    # in the atmosphere.

    X = 0.02 * (t - 12.5 + 7500. / p)
    WFW = 1. + 4.5e-06 * p + 1.4e-03 * X * X
    FWESW = WFW * esw_skewt_t(t)
    R = EPS * FWESW / (p - FWESW)

    # convert r from a dimensionless ratio to grams/kilogram.

    return 1000. * R


def esw_skewt_t(t):
    """
    Arguments:
        t - temperature in degrees C
    Returns:
        esw - Saturation vapor pressure in millibars
    Adapted from NCL
    this function returns the saturation vapor pressure esw (millibars)
    over liquid water given the temperature t (celsius). the polynomial
    approximation below is due to herman wobus, a mathematician who
    worked at the navy weather research facility, norfolk, virginia,
    but who is now retired. the coefficients of the polynomial were
    chosen to fit the values in table 94 on pp. 351-353 of the smith-
    sonian meteorological tables by roland list (6th edition). the
    approximation is valid for -50 < t < 100c.

    es0 = saturation vapor ressure over liquid water at 0c
    """
    ES0 = 6.1078e0

    POL = (0.99999683e0 + t * (-0.90826951e-02 +
                               t *
                               (0.78736169e-04 +
                                t *
                                (-0.61117958e-06 +
                                 t *
                                 (0.43884187e-08 +
                                  t * (-0.29883885e-10 +
                                       t *
                                       (0.21874425e-12 +
                                        t *
                                        (-0.17892321e-14 +
                                         t * (0.11112018e-16 +
                                              t * (-0.30994571e-19)
                                              )
                                         )
                                        )
                                       )
                                  )
                                 )
                                )
                               )
           )
    return ES0 / POL**8


def uv_wswd(ws, wd, isradians=False):
    """
    Arguments:
        ws - wind speed array
        wd - wind direction array (in degrees unless isradians is set to true)
        isradians - False if WD is in degrees
    Returns:
        U, V - u-component and v-component winds

    https://www.eol.ucar.edu/projects/ceop/dm/documents/
    refdata_report/eqns.html
    """
    dimensions = getattr(ws, 'dimensions', getattr(
        wd, 'dimensions', ('unknown',)))
    units = getattr(ws, 'units', 'unknown')
    if isradians:
        direction = wd
    else:
        direction = np.radians(wd)
    wind_speed = ws
    U = -np.sin(direction) * wind_speed
    V = -np.cos(direction) * wind_speed
    u = PseudoNetCDFVariable(None, 'U', U.dtype.char,
                             dimensions, values=U, units=units, short_name='U')
    v = PseudoNetCDFVariable(None, 'V', V.dtype.char,
                             dimensions, values=V, units=units, short_name='V')
    return u, v


def wswd_uv(u, v, return_radians=False):
    """
    Arguments:
        u - array of u-component winds
        v - array of v-domponent winds
        return_radians - convert direction to radians
    Returns:
        ws, wd - wind speed and direction (from direction)

    https://www.eol.ucar.edu/projects/ceop/dm/documents/refdata_report/
    eqns.html
    """
    dimensions = getattr(u, 'dimensions', getattr(
        v, 'dimensions', ('unknown',)))
    units = getattr(u, 'units', getattr(v, 'units', 'unknown'))
    wind_speed = np.sqrt(u**2 + v**2)
    uvarctan = np.arctan(u / v)
    wind_direction = np.where(v < 0, uvarctan * 180. /
                              np.pi, uvarctan * 180. / np.pi + 180.)
    if return_radians:
        wind_direction = np.radians(wind_direction)
    ws = PseudoNetCDFVariable(None, 'WS', wind_speed.dtype.char,
                              dimensions, values=wind_speed,
                              units=units, short_name='WS')
    dirunit = 'radians' if return_radians else 'degrees'
    wd = PseudoNetCDFVariable(None, 'WD', wind_speed.dtype.char, dimensions,
                              values=wind_direction,
                              units=dirunit, short_name='WD')
    return ws, wd



__all__ = ['ipr']
__doc__ = """
.. _Read
:mod:`Read` -- ipr Read interface
============================================

.. module:: Read
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` random access read for CAMx
              ipr files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
# Distribution packages
import unittest
import struct
from warnings import warn

# Site-Packages
from numpy import zeros, dtype, fromfile
from numpy import char

# This Package modules
from PseudoNetCDF.conventions.ioapi import add_cf_from_ioapi
from PseudoNetCDF.camxfiles.timetuple import timediff
from PseudoNetCDF.camxfiles.util import cartesian
from PseudoNetCDF.camxfiles.units import get_uamiv_units
from PseudoNetCDF.sci_var import PseudoNetCDFFile
from PseudoNetCDF.sci_var import PseudoNetCDFVariable as pncvar
from PseudoNetCDF.sci_var import PseudoNetCDFVariables
from PseudoNetCDF.ArrayTransforms import ConvertCAMxTime

# for use in identifying uncaught nan
listnan = struct.unpack('>f', b'\xff\xc0\x00\x00')[0]
checkarray = zeros((1,), 'f')
checkarray[0] = listnan
array_nan = checkarray[0]


class ipr(PseudoNetCDFFile):
    """
    ipr provides a PseudoNetCDF interface for CAMx
    ipr files.  Where possible, the inteface follows
    IOAPI conventions (see www.baronams.com).

    ex:
        >>> ipr_path = 'camx_ipr.bin'
        >>> rows,cols = 65,83
        >>> iprfile = ipr(ipr_path,rows,cols)
        >>> iprfile.variables.keys()
        ['TFLAG', 'SPAD_O3', 'DATE_O3', 'TIME_O3', 'SPC_O3',
         'PAGRID_O3', 'NEST_O3', 'I_O3', 'J_O3', 'K_O3',
         'INIT_O3', 'CHEM_O3', 'EMIS_O3', 'PTEMIS_O3',
         'PIG_O3', 'WADV_O3', 'EADV_O3', 'SADV_O3', 'NADV_O3',
         'BADV_O3', 'TADV_O3', 'DIL_O3', 'WDIF_O3', 'EDIF_O3',
         'SDIF_O3', 'NDIF_O3', 'BDIF_O3', 'TDIF_O3', 'DDEP_O3',
         'WDEP_O3', 'INORGACHEM_O3', 'ORGACHEM_O3', 'AQACHEM_O3',
         'FCONC_O3', 'UCNV_O3', 'AVOL_O3', 'EPAD_O3']
        >>> v = iprfile.variables['CHEM_O3']
        >>> tflag = iprfile.variables['TFLAG']
        >>> tflag.dimensions
        ('TSTEP', 'VAR', 'DATE-TIME')
        >>> tflag[0,0,:]
        array([2005185,       0])
        >>> tflag[-1,0,:]
        array([2005185,  240000])
        >>> v.dimensions
        ('TSTEP', 'LAY', 'ROW', 'COL')
        >>> v.shape
        (25, 28, 65, 83)
        >>> iprfile.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 65, 'COL': 83}
    """

    __ipr_record_type = {
        24: dtype(
            dict(
                names=['SPAD', 'DATE', 'TIME', 'SPC', 'PAGRID', 'NEST', 'I',
                       'J', 'K', 'INIT', 'CHEM', 'EMIS', 'PTEMIS', 'PIG',
                       'WADV', 'EADV', 'SADV', 'NADV', 'BADV', 'TADV', 'DIL',
                       'WDIF', 'EDIF', 'SDIF', 'NDIF', 'BDIF', 'TDIF', 'DDEP',
                       'WDEP', 'AERCHEM', 'FCONC', 'UCNV', 'AVOL', 'EPAD'],
                formats=['>i', '>i', '>f', '>S10', '>i', '>i', '>i', '>i',
                         '>i', '>f', '>f', '>f', '>f', '>f', '>f', '>f', '>f',
                         '>f', '>f', '>f', '>f', '>f', '>f', '>f', '>f', '>f',
                         '>f', '>f', '>f', '>f', '>f', '>f', '>f', '>i']
            )
        ),
        26: dtype(
            dict(
                names=['SPAD', 'DATE', 'TIME', 'SPC', 'PAGRID', 'NEST', 'I',
                       'J', 'K', 'INIT', 'CHEM', 'EMIS', 'PTEMIS', 'PIG',
                       'WADV', 'EADV', 'SADV', 'NADV', 'BADV', 'TADV', 'DIL',
                       'WDIF', 'EDIF', 'SDIF', 'NDIF', 'BDIF', 'TDIF', 'DDEP',
                       'WDEP', 'INORGACHEM', 'ORGACHEM', 'AQACHEM', 'FCONC',
                       'UCNV', 'AVOL', 'EPAD'],
                formats=['>i', '>i', '>f', '>S10', '>i', '>i', '>i', '>i',
                         '>i', '>f', '>f', '>f', '>f', '>f', '>f', '>f', '>f',
                         '>f', '>f', '>f', '>f', '>f', '>f', '>f', '>f', '>f',
                         '>f', '>f', '>f', '>f', '>f', '>f', '>f', '>f', '>f',
                         '>i']
            )
        )
    }

    def __init__(self, rf, proc_dict=None, units='umol/m**3',
                 oldnames=False, **props):
        """
        Initialization included reading the header and learning
        about the format.

        see __readheader and __gettimestep() for more info

        Keywords (i.e., props) for projection: P_ALP, P_BET, P_GAM, XCENT,
            YCENT, XORIG, YORIG, XCELL, YCELL
        """
        if proc_dict is not None:
            self.proc_dict = proc_dict
        else:
            self.proc_dict = None
        self.__rffile = open(rf, 'rb')
        self.__rffile.seek(0, 2)
        if self.__rffile.tell() < 2147483648:
            warn("For greater speed on files <2GB use ipr_memmap")
        self.__rffile.seek(0, 0)
        self.units = units
        self.__readheader()
        self.__setDomain__()
        self.__gettimestep()
        tdim = self.createDimension('TSTEP', self.NSTEPS)
        tdim.setunlimited(True)
        self.createDimension('DATE-TIME', 2)
        self.createDimension('VAR', self.NSPCS * self.NPROCESS)
        spcstrs = char.decode(self.spcnames['SPECIES']).tolist()
        spcstrs = [s.strip() for s in spcstrs]
        prcstrs = self.proc_dict.keys()
        varkeys = ["_".join([j[1], j[0]]) for j in cartesian(spcstrs, prcstrs)]
        varkeys += ['TFLAG']
        self.variables = PseudoNetCDFVariables(self.__variables, varkeys)
        for k, v in props.items():
            setattr(self, k, v)
        try:
            add_cf_from_ioapi(self)
        except Exception:
            pass

    def __variables(self, proc_spc):
        if proc_spc == 'TFLAG':
            time = self.variables['TIME_%s' %
                                  char.decode(self.spcnames)[0][1].strip()]
            date = self.variables['DATE_%s' %
                                  char.decode(self.spcnames[0])[1].strip()]
            tmpvals = ConvertCAMxTime(date[:, 0, 0, 0],
                                      time[:, 0, 0, 0],
                                      len(self.dimensions['VAR']))
            tmpvar = pncvar(self, 'proc_spc', 'i',
                            ('TSTEP', 'VAR', 'DATE-TIME'),
                            values=tmpvals)
            self.variables['TFLAG'] = tmpvar
            return self.variables['TFLAG']

        self.variables.clear()

        spcstrs = char.decode(self.spcnames['SPECIES']).tolist()
        for k in self.proc_dict:
            proc = proc_spc[:len(k)]
            spc = proc_spc[len(k) + 1:]
            if proc == k and spc.ljust(10) in spcstrs:
                spcprocs = self.__readalltime(spc)
                for p, plong in self.proc_dict.items():
                    var_name = p + '_' + spc
                    # IPR units are consistent with 'IPR'
                    if p == 'UCNV':
                        units = 'm**3/mol'
                    elif p == 'AVOL':
                        units = 'm**3'
                    else:
                        units = get_uamiv_units('IPR', spc)
                    tmpvar = pncvar(self, var_name, 'f',
                                    ('TSTEP', 'LAY', 'ROW', 'COL'),
                                    values=spcprocs[p],
                                    units=units,
                                    var_desc=var_name.ljust(16),
                                    long_name=var_name.ljust(16))
                    self.variables[var_name] = tmpvar
                del spcprocs
                return self.variables[proc_spc]
        raise KeyError("Bad!")

    def __readonetime(self, ntime, spc):
        newpos = self.__start(ntime, spc)
        oldpos = self.__rffile.tell()
        relmove = newpos - oldpos
        self.__rffile.seek(relmove, 1)
        tmpout = fromfile(self.__rffile, dtype=self.__ipr_record_type,
                          count=self.__block3d)

        return tmpout.reshape(self.NROWS, self.NCOLS, self.NLAYS)\
                     .swapaxes(0, 2)\
                     .swapaxes(1, 2)

    def __readalltime(self, spc):
        out = zeros((self.NSTEPS, self.NLAYS, self.NROWS,
                     self.NCOLS), dtype=self.__ipr_record_type)
        for it in range(self.NSTEPS):
            out[it] = self.__readonetime(it, spc)
        return out

    def __start(self, ntime, spc):
        nspec = char.decode(
            self.spcnames['SPECIES']).tolist().index(spc.ljust(10))
        return (self.__data_start_byte +
                (int(ntime) * self.__block4d + self.__block3d * nspec) *
                self.__ipr_record_type.itemsize)

    def __readheader(self):
        """
        __readheader reads the header section of the ipr file
        it initializes each header field (see CAMx Users Manual for a list)
        as properties of the ipr class
        """

        self.runmessage = fromfile(self.__rffile, dtype=dtype(dict(
            names=['SPAD', 'RUNMESSAGE', 'EPAD'],
            formats=['>i', '>80S', '>i'])), count=1)['RUNMESSAGE']
        dates = fromfile(self.__rffile, dtype=dtype(
            dict(names=['SPAD', 'SDATE', 'STIME', 'EDATE', 'ETIME', 'EPAD'],
                 formats=['>i', '>i', '>f', '>i', '>f', '>i'])), count=1)
        self.SDATE = dates['SDATE'] + 2000000
        self.STIME = dates['STIME']
        self.EDATE = dates['EDATE'] + 2000000
        self.ETIME = dates['ETIME']

        self.__grids = []
        self.NGRIDS = fromfile(self.__rffile, dtype=dtype(
            dict(names=['SPAD', 'NGRIDS', 'EPAD'],
                 formats=['>i'] * 3)), count=1)['NGRIDS']
        for grid in range(self.NGRIDS):
            gddt = dtype(dict(names=['SPAD', 'orgx', 'orgy', 'ncol', 'nrow',
                                     'xsize', 'ysize', 'EPAD'],
                              formats=['>i', '>i', '>i', '>i', '>i', '>i',
                                       '>i', '>i']))
            self.__grids.append(
                fromfile(self.__rffile, dtype=gddt, count=1)
            )

        self.spcnames = []
        self.NSPCS = fromfile(self.__rffile, dtype=dtype(dict(
            names=['SPAD', 'NSPCS', 'EPAD'],
            formats=['>i', '>i', '>i'])), count=1)['NSPCS'][0]
        self.spcnames = fromfile(self.__rffile, dtype=dtype(dict(
            names=['SPAD', 'SPECIES', 'EPAD'],
            formats=['>i', '>10S', '>i'])), count=self.NSPCS)

        self.padomains = []
        self.NPADOMAINS = fromfile(self.__rffile, dtype=dtype(dict(
            names=['SPAD', 'NPADOMAINS', 'EPAD'],
            formats=['>i', '>i', '>i'])), count=1)['NPADOMAINS'][0]
        self.__padomains = fromfile(
            self.__rffile,
            dtype=dtype(dict(
                names=['SPAD', 'grid', 'istart', 'iend', 'jstart', 'jend',
                       'blay', 'tlay', 'EPAD'],
                formats=['>i', '>i', '>i', '>i', '>i', '>i', '>i', '>i',
                         '>i']
            )),
            count=self.NPADOMAINS
        )
        self.__activedomain = self.__padomains[0]
        self.prcnames = []
        self.NPROCESS = fromfile(self.__rffile, dtype=dtype(dict(
            names=['SPAD', 'NPRCS', 'EPAD'],
            formats=['>i', '>i', '>i'])), count=1)['NPRCS']

        self.__ipr_record_type = self.__ipr_record_type[self.NPROCESS[0]]

        if self.proc_dict is None:
            self.proc_dict = {
                'INIT': 'Initial concentration',
                'CHEM': 'Chemistry',
                'EMIS': 'Area emissions',
                'PTEMIS': 'Point source emissions',
                'PIG': 'Plume-in-Grid change',
                'WADV': 'West boundary advection',
                'EADV': 'East boundary advection',
                'SADV': 'South boundary advection',
                'NADV': 'North boundary advection',
                'BADV': 'Bottom boundary advection',
                'TADV': 'Top boundary advection',
                'DIL': 'Dilution in the vertical',
                'WDIF': 'West boundary diffusion',
                'EDIF': 'East boundary diffusion',
                'SDIF': 'South boundary diffusion',
                'NDIF': 'North boundary diffusion',
                'BDIF': 'Bottom boundary diffusion',
                'TDIF': 'Top boundary diffusion',
                'DDEP': 'Dry deposition',
                'WDEP': 'Wet deposition',
                'FCONC': 'Final concentration',
                'UCNV': 'Units conversion',
                'AVOL': 'Average cell volume',
                'DATE': 'DATE',
                'TIME': 'TIME',
                'K': 'K',
                'J': 'J',
                'I': 'I'
            }
            if self.NPROCESS[0] == 24:
                self.proc_dict['AERCHEM'] = 'Aerosol chemistry'
            elif self.NPROCESS[0] == 26:
                self.proc_dict['INORGACHEM'] = 'Inorganic Aerosol chemistry'
                self.proc_dict['ORGACHEM'] = 'Organic Aerosol chemistry'
                self.proc_dict['AQACHEM'] = 'Aqueous Aerosol chemistry'
            else:
                warn('Unknown version; cannot add aerosol chemistry')

        prdt = dtype(dict(names=['SPAD', 'PROCESS', 'EPAD'],
                          formats=['>i', '>25S', '>i']))
        self.prcnames = fromfile(self.__rffile, dtype=prdt,
                                 count=self.NPROCESS)
        self.__data_start_byte = self.__rffile.tell()

    def __setDomain__(self, id=0):
        self.__activedomain = self.__padomains[id]
        ncols = self.__activedomain['iend'] - self.__activedomain['istart'] + 1
        self.createDimension('COL', ncols)
        nrows = self.__activedomain['jend'] - self.__activedomain['jstart'] + 1
        self.createDimension('ROW', nrows)
        nlays = self.__activedomain['tlay'] - self.__activedomain['blay'] + 1
        self.createDimension('LAY', nlays)
        self.NCOLS = len(self.dimensions['COL'])
        self.NROWS = len(self.dimensions['ROW'])
        self.NLAYS = len(self.dimensions['LAY'])
        self.__block3d = self.NLAYS * self.NROWS * self.NCOLS
        self.__block4d = self.__block3d * self.NSPCS

    def __gettimestep(self):
        """
        Header information provides start and end date, but does not
        indicate the increment between.  This routine reads the first
        and second date/time and initializes variables indicating the
        timestep length and the anticipated number.
        """
        self.__rffile.seek(self.__data_start_byte, 0)
        rcount = (len(self.dimensions['LAY']) *
                  len(self.dimensions['ROW']) *
                  len(self.dimensions['COL']) + 1)
        temp = fromfile(self.__rffile,
                        dtype=self.__ipr_record_type,
                        count=rcount)
        self.TSTEP = timediff((self.SDATE, self.STIME),
                              (temp[-1]['DATE'] + 2000000, temp[-1]['TIME']))
        self.NSTEPS = int(timediff((self.SDATE, self.STIME),
                                   (self.EDATE, self.ETIME)) / self.TSTEP)


class TestRead(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testIPR(self):
        warn('Test not implemented; data too big for distribution')


if __name__ == '__main__':
    unittest.main()



import os
import re
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from PseudoNetCDF.geoschemfiles import bpch


def cspec(path, smv2='smv2.log'):
    if not os.path.isfile(smv2):
        if os.path.isdir(smv2):
            smv2 = os.path.join(smv2, 'smv2.log')
        else:
            smv2 = os.path.join(os.path.dirname(path), smv2)
    diaginfo, tracerinfo = get_info_for_cspec(get_cspec(smv2))
    return bpch(path, diaginfo=diaginfo, tracerinfo=tracerinfo)


def get_cspec(smvpath):
    text = open(smvpath, 'r').read()
    reo = re.compile('(?:NBR NAME\s+MW BKGAS\(VMRAT\))(?P<active>.*)' +
                     '(?:INACTIVE SPECIES FOR THIS RUN ARE:)' +
                     '(?P<inactive>.*)(?:THE DEAD SPECIES FOR THIS RUN ARE:)' +
                     '(?P<dead>.*?)(?:=====)', re.M | re.DOTALL)
    gv = reo.search(text).groupdict()
    active = [v.strip().split()[1] for v in gv['active'].strip().split('\n')]
    inactive = gv['inactive'].replace('\n', ' ').split()
    # dead = gv['dead'].replace('\n', ' ').split()
    all = active + inactive  # + dead
    return all


def get_info_for_cspec(all):
    out = ""
    for si, spc in enumerate(all):
        out += "%-8s %-8s cspec                 " % (spc, spc)
        out += "0.000E-00  1 %-8d 1.000E+00 molec/cm3\n" % (si + 1)

    initstr = ("#\n       0 IJ-CHK-$                                 " +
               "Tracer concentration                    \n")
    diag = StringIO(initstr)
    return diag, StringIO(out)


if __name__ == '__main__':
    specs = get_cspec('testdata/smv2.log')
    diagfile, trcfile = get_info_for_cspec(specs)
    f = bpch('/Users/barronh/Development/testrun/BC.CSPEC.20050101',
             tracerinfo=trcfile, diaginfo=diagfile)



__all__ = ['timerange', 'timediff', 'timeadd', 'cmp_time']

__doc__ = """
.. _timetuple
:mod:`timetuple` -- CAMx simple time functions
==============================================

.. module:: timetuple
   :platform: Unix, Windows
   :synopsis: Time tuple provided a simple interface for handling
              CAMx time representations.  It has been deprecated
              and all reliance should be removed
.. moduleauthor:: Adam Hupp and Barron Henderson <barronh@unc.edu>
"""

import unittest


def timerange(datetime1, datetime2, step=100, eod=2400.0):
    """Iterater of time tuples between start and end (not end inclusive)

    Should probably remove the != and use < if step + and > if step -
    """
    (date1, time1) = datetime1
    (date2, time2) = datetime2
    date1, time1 = timeadd((date1, time1), (0, 0), eod)
    date2, time2 = timeadd((date2, time2), (0, 0), eod)
    while (date1, time1) != (date2, time2):
        yield date1, time1
        date1, time1 = timeadd((date1, time1), (0, step), eod)


def timediff(datetime1, datetime2, eod=2400.0):
    """Compares date tuples and returns difference in
    a time tuple
    """
    (date1, time1) = datetime1
    (date2, time2) = datetime2
    date3 = date2 - date1
    time3 = time2 - time1
    return date3 * eod + time3


def timeadd(datetime1, datetime2, eod=2400.0):
    """Adds time tuple 1 to time tuple 2.
    Time values greater than eod (dflt 2400)
    are converted to days

    Big question is hr 24 day 1
    """
    (date1, time1) = datetime1
    (date2, time2) = datetime2

    time1 = time1 + time2
    date1 = date1 + date2
    if time1 >= eod:
        time1 %= eod
        date1 += 1
    if time1 < 0:
        time1 %= eod
        date1 -= 1
    return date1, time1


def cmp_time(lhs, rhs):
    """Compare two time tuples.
    Each arg is a 2-tuple, the first element being the hour of the day * 100,
    the second being the date
    """
    ltime, ldate = lhs
    rtime, rdate = rhs
    if ldate == rdate:
        return ltime - rtime
    else:
        return ldate - rdate


#
# Test cases
#

class CompareTime(unittest.TestCase):
    def testCompare(self):
        assert cmp_time((1.0, 2), (2.0, 2)) < 0
        assert cmp_time((2.0, 2), (2.0, 2)) == 0
        assert cmp_time((3.0, 2), (2.0, 3)) < 0
        assert cmp_time((3.0, 3), (2.0, 3)) > 0



__all__ = ['PseudoNetCDFFile', 'netcdf', 'PseudoNetCDFVariables']
import unittest
from PseudoNetCDF._getreader import registerreader
from PseudoNetCDF.netcdf import NetCDFFile, NetCDFVariable
from PseudoNetCDF.pncwarn import warn
from collections import OrderedDict
from ._dimensions import PseudoNetCDFDimension, PseudoNetCDFDimensions
from ._variables import PseudoNetCDFVariable, PseudoNetCDFMaskedVariable
import numpy as np


class OrderedDefaultDict(OrderedDict):
    def __init__(self, *args, **kwargs):
        if not args:
            self.default_factory = None
        else:
            if not (args[0] is None or callable(args[0])):
                raise TypeError('first argument must be callable or None')
            self.default_factory = args[0]
            args = args[1:]
        super(OrderedDefaultDict, self).__init__(*args, **kwargs)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = default = self.default_factory()
        return default


class PseudoNetCDFType(type):
    """
    Create a PseudoNetCDFType meta-class
    """
    def __init__(cls, name, bases, clsdict):
        pieces = str(cls).split('\'')[1].split('.')
        longname = '.'.join(
            [p for p in pieces[1:-1] if '_' != p[0] and p not in ('core',)] +
            [pieces[-1]])
        if len(cls.mro()) > 2:
            if name not in ('PseudoNetCDFFile', 'WrapPnc'):
                shortl = registerreader(name, cls)
                longl = registerreader(longname, cls)
                if not (shortl or longl):
                    warn('Not registered either as ' + name +
                         ' or ' + longname)
        super(PseudoNetCDFType, cls).__init__(name, bases, clsdict)


PseudoNetCDFSelfReg = PseudoNetCDFType('pnc', (object,), dict(__doc__='Test'))


class PseudoNetCDFFile(PseudoNetCDFSelfReg, object):
    """
    PseudoNetCDFFile provides an interface and standard set of
    methods that a file should present to act like a netCDF file
    using the Scientific.IO.NetCDF.NetCDFFile interface.
    """

    def getMap(self, maptype='basemap_auto', **kwds):
        """
        Description

        Parameters
        ----------
        maptype : string
            choices 'basemap', 'basemap_auto', 'cartopy' (not yet)
            basemap : attempts to open a basemap with only supplied kwds
            basemap_auto : automatically adds llcrnrlon,llcrnrlat,u
                           rcrnrlon,urcrnrlat based on longitude_bounds
        **kwds : keywords
            for basemap or cartopy

        Returns
        -------
        map : basemap or cartopy axis
        """
        if maptype.startswith('basemap'):
            from PseudoNetCDF.coordutil import basemap_from_proj4
            if (
                maptype.endswith('_auto') and
                'longitude_bounds' in self.variables and
                'latitude_bounds' in self.variables
            ):
                # Get edges for bounding
                lonb = self.variables['longitude_bounds']
                latb = self.variables['latitude_bounds']
                if lonb.ndim == 3:
                    llcrnrlon = lonb[0, 0, 0]
                    urcrnrlon = lonb[-1, -1, 2]
                elif lonb.ndim == 2:
                    llcrnrlon = lonb[0, 0]
                    urcrnrlon = lonb[-1, -1]
                elif lonb.ndim == 1:
                    llcrnrlon = lonb[0]
                    urcrnrlon = lonb[-1]
                if latb.ndim == 3:
                    llcrnrlat = latb[0, 0, 0]
                    urcrnrlat = latb[-1, -1, 2]
                elif latb.ndim == 2:
                    llcrnrlat = latb[0, 0]
                    urcrnrlat = latb[-1, -1]
                elif latb.ndim == 1:
                    llcrnrlat = latb[0]
                    urcrnrlat = latb[-1]
                edges = dict(llcrnrlon=llcrnrlon, llcrnrlat=llcrnrlat,
                             urcrnrlon=urcrnrlon, urcrnrlat=urcrnrlat)
                kwds.update(edges)
            myproj = self.getproj(withgrid=True, projformat='proj4')
            return basemap_from_proj4(myproj, **kwds)
        elif maptype == 'cartopy':
            raise ValueError('cartopy is not yet implemented')
        else:
            raise ValueError(
                'maptype must be basemap, basemap_auto, or cartopy')

    def getproj(self, withgrid=False, projformat='pyproj'):
        """
        Description

        Parameters
        ----------
        withgrid : boolean
            use grid units instead of meters
        projformat : string
            'pyproj' (default), 'proj4' or 'wkt' allows function to
            return a pyproj projection object or a string in the
            format of proj4 or WKT

        Returns
        -------
        proj : string pyproj.Proj
             (wkt, proj4) or pyprojProj (pyproj)
        """
        if projformat == 'pyproj':
            from PseudoNetCDF.coordutil import getproj
            return getproj(self, withgrid=withgrid)
        elif projformat == 'proj4':
            from PseudoNetCDF.coordutil import getproj4
            return getproj4(self, withgrid=withgrid)
        elif projformat == 'wkt':
            from PseudoNetCDF.coordutil import getprojwkt
            return getprojwkt(self, withgrid=withgrid)
        else:
            raise ValueError('projformat must be pyproj, proj4 or wkt')

    def ll2xy(self, lon, lat):
        """
        Converts lon/lat to x distances (no false easting/northing)

        Parameters
        ----------
        lon : scalar or iterable
            longitudes in decimal degrees
        lat : scalar or iterable
            latitudes in decimal degrees

        Returns
        -------
        x, y : tuple of arrays
            coordinates in map projection (meters or radians)
        """
        lon = np.asarray(lon)
        lat = np.asarray(lat)
        return self.getproj()(lon, lat)

    def _getzdim(self):
        """
        Heuristic function finds and returns the name of the vertical dimension
        based on common dimension names. Overwrite method for explicit control.

        Returns
        -------
        result : {'LAY', 'bottom_top', 'level', 'lev', 'layer', 'lay', 'z'}

        Examples
        --------
        >>> f = pnc.pncopen('', format='PseudoNetCDFFile')
        >>> f.createDimension('t', 4)
        >>> f.createDimension('x', 4)
        >>> f.createDimension('y', 4)
        >>> f.createDimension('z', 4)
        >>> f._getzdim()
        'z'
        """
        for dk in 'LAY bottom_top level lev layer lay z'.split():
            if dk in self.dimensions:
                return dk
        else:
            raise KeyError('Could not find z dimensions')

    def _gettdim(self):
        """
        Heuristic function finds and returns the name of the time dimension
        based on common dimension names. Overwrite method for explicit control.

        Returns
        -------
        result : {'time', 'TSTEP', 't', 'Time'}

        Examples
        --------
        >>> f = pnc.pncopen('', format='PseudoNetCDFFile')
        >>> f.createDimension('t', 4)
        >>> f.createDimension('x', 4)
        >>> f.createDimension('y', 4)
        >>> f.createDimension('z', 4)
        >>> f._gettdim()
        't'
        """
        for dk in 'time TSTEP t Time'.split():
            if dk in self.dimensions:
                return dk
        else:
            raise KeyError('Could not find t dimensions')

    def _getydim(self):
        """
        Heuristic function finds and returns the name of the y or width
        dimension based on common dimension names. Overwrite method for
        explicit control.

        Returns
        -------
        result : {'latitude', 'lat', 'south_north', 'ROW', 'y'}

        Examples
        --------
        >>> f = pnc.pncopen('', format='PseudoNetCDFFile')
        >>> f.createDimension('t', 4)
        >>> f.createDimension('x', 4)
        >>> f.createDimension('y', 4)
        >>> f.createDimension('z', 4)
        >>> f._getydim()
        'y'
        """
        for dk in 'latitude lat south_north ROW y'.split():
            if dk in self.dimensions:
                return dk
        else:
            raise KeyError('Could not find y dimensions')

    def _getxdim(self):
        """
        Heuristic function finds and returns the name of the x or height
        dimension based on common dimension names. Overwrite method for
        explicit control.

        Returns
        -------
        result : {'longitude', 'lon', 'west_east', 'COL', 'x'}

        Examples
        --------
        >>> f = pnc.pncopen('', format='PseudoNetCDFFile')
        >>> f.createDimension('t', 4)
        >>> f.createDimension('x', 4)
        >>> f.createDimension('y', 4)
        >>> f.createDimension('z', 4)
        >>> f._getydim()
        'y'
        """
        for dk in 'longitude long lon west_east COL x'.split():
            if dk in self.dimensions:
                return dk
        else:
            raise KeyError('Could not find x dimensions')

    def ll2ij(self, lon, lat, bounds='ignore', clean='none'):
        """
        Converts lon/lat to 0-based indicies (0,M), (0,N)

        Parameters
        ----------
        lon : scalar or iterable
            longitudes in decimal degrees
        lat : scalar or iterable
            latitudes in decimal degrees
        bounds : string
            ignore, error, warn if i,j are out of domain
        clean : string
            none - return values regardless of bounds;
            mask - mask values out of bounds;
            clip - return min(max(0, v), nx - 1)

        Returns
        -------
        i, j : indices (0-based) for variables
        """
        lon = np.asarray(lon)
        lat = np.asarray(lat)
        p = self.getproj(withgrid=True)
        x, y = p(lon, lat)
        i = np.asarray(x).astype('i')
        j = np.asarray(y).astype('i')
        nx = len(self.dimensions[self._getxdim()])
        ny = len(self.dimensions[self._getydim()])
        if bounds == 'ignore':
            pass
        else:
            lowi = (i < 0)
            lowj = (j < 0)
            highi = (i >= nx)
            highj = (j >= ny)
            outb = (lowi | lowj | highi | highj)
            nout = outb.sum()
            if nout > 0:
                message = '{} Points out of bounds; {}'.format(
                    nout, np.where(outb))
                if bounds == 'error':
                    raise ValueError(message)
                else:
                    warn(message)
        if clean == 'clip':
            i = np.minimum(np.maximum(i, 0), nx - 1)
            j = np.minimum(np.maximum(j, 0), ny - 1)
        elif clean == 'mask':
            i = np.ma.masked_greater(np.ma.masked_less(i, 0), nx - 1)
            j = np.ma.masked_greater(np.ma.masked_less(j, 0), ny - 1)

        return i, j

    def xy2ll(self, x, y):
        """
        Converts x, y to lon, lat (no false easting/northing)

        Parameters
        ----------
        x : scalar or iterable
            projected west-east coordinates
        y : scalar or iterable
            projected south-north coordinates

        Returns
        -------
        lon, lat : scalars or iterables
            longitudes and latitudes in decimal degrees
        """
        x = np.asarray(x)
        y = np.asarray(y)
        p = self.getproj()
        lon, lat = p(x, y, inverse=True)
        return lon, lat

    def ij2ll(self, i, j):
        """
        Converts i, j to lon, lat (no false easting/northing)
        using cell centers assuming 0-based i/j

        Parameters
        ----------
        i : scalar/iterable
            indicies (0-based) for the west-east dimension
        j : scalar/iterable
            indicies (0-based) for the south-north dimension

        Returns
        -------
        lon, lat : scalars or iterables
            longitudes and latitudes in decimal degrees
        """
        i = np.asarray(i)
        j = np.asarray(j)
        p = self.getproj(withgrid=True)
        lon, lat = p(i + 0.5, j + 0.5, inverse=True)
        return lon, lat

    def date2num(self, time, timekey='time'):
        """
        Parameters
        ----------
        time : array-like
            array of datetime.datetime objects
        timekey : str
            time variable key which requires units and should have calendar.
            If calendar is missing, standard is the default. default 'time'

        Returns
        -------
        num : array-like
            time in relative time as defined by units of time variable
            (i.e., timekey) which defaults to 'time'
        """
        from netCDF4 import date2num
        try:
            from datetime import timezone
            utc = timezone.utc
        except ImportError:
            from datetime import tzinfo
            utc = tzinfo.utc

        time = np.asarray(time)
        # netCDF4 date2num is timezone naive; assumes UTC when not
        # specified and converts to UTC internally
        # so, if a tzinfo is involved, it should be removed
        if any([t.tzinfo is not None for t in time[:]]):
            time = np.array([
                t.astimezone(utc).replace(tzinfo=None) for t in time[:]
            ])

        timeunits = self.variables[timekey].units.strip()
        calendar = getattr(self.variables[timekey], 'calendar', 'standard')
        num = date2num(time, timeunits, calendar.strip())
        return num

    def time2idx(self, time, dim='time', timekey=None, **kwds):
        """
        Convert datetime objects to dimension indices

        Parameters
        ----------
        time : array-like
            array of datetime.datetime objects
        dim : str
            dimension name for val2idx
        timekey : str
            time variable key. None defaults to dim
        kwds : mappable
            see val2idx

        Returns
        -------
        idx : array-like
            time index (0-based)
        """
        if timekey is None:
            timekey = dim

        time = np.asarray(time)
        nums = self.date2num(time, timekey=timekey)
        return self.val2idx(dim=dim, val=nums, **kwds)

    def time2t(self, time, ttype='nearest', index=True):
        """
        Parameters
        ----------
        time : array of datetime.datetime objects
        interp : 'nearest', 'bounds', 'bounds_close'
        index : return index

        Returns
        -------
        t : fractional time or if index, integers for indexing
        """
        warn(
            "time2t is deprecated; transition to time2idx or date2num",
            DeprecationWarning
        )
        time = np.asarray(time)
        if ttype not in ('nearest', 'bounds', 'bounds_close'):
            warn('{} is not an option; defaulting to nearest'.format(ttype))
            ttype = 'nearest'
        if ttype == 'nearest':
            mytimes = self.getTimes()
        else:
            mytimes = self.getTimes(bounds=True)
        idx = np.arange(mytimes.size)

        # Find minimum resoultion
        for res in [
            'microsecond', 'second', 'minute', 'hour', 'day', 'month', 'year'
        ]:
            minres = res
            resvals = [getattr(d, res) for d in time]
            myresvals = [getattr(d, res) for d in mytimes]
            if not np.allclose(myresvals, 0):
                break
            if not np.allclose(resvals, 0):
                break

        # Translate minimum resolution to numpy datetime
        if minres == 'microsecond':
            tu = 'datetime64[ns]'
        elif minres == 'second':
            tu = 'datetime64[s]'
        elif minres == 'minute':
            tu = 'datetime64[m]'
        elif minres == 'hour':
            tu = 'datetime64[h]'
        elif minres in ('year', 'month', 'day'):
            tu = 'datetime64[D]'
        else:
            tu = 'datetime64[ns]'

        # Convert input time to numpy datetime at resolution
        x = time.astype(tu).astype('d')
        # Convert file's time to numpy datetime at resolution
        xp = mytimes.astype(tu).astype('d')

        # Use interpolation methods with no bounding for nearest
        # and bounds_close
        if ttype in ('nearest', 'bounds_close'):
            out = np.interp(x, xp, idx)
            if index:
                imin = 0
                imax = idx[-1] + (0 if ttype == 'nearest' else -1)
                out = np.minimum(np.maximum(out, imin), imax)
                if ttype == 'nearest':
                    out = np.round(out, 0).astype('i')
                else:
                    out = np.floor(out).astype('i')
        # Use interpolation methods with bounding for nearest
        else:
            out = np.interp(x, xp, idx, left=np.nan, right=np.nan)
            if index:
                out = np.ma.masked_less(np.ma.floor(out).astype('i'), 0)

        return out

    def val2idx(
        self, dim, val,
        method='nearest', bounds='warn', left=None, right=None, clean='mask'
    ):
        """
        Convert coordinate values to indices

        Parameters
        ----------
        dim : str
            name of dimensions, which must have a coordinate variable
        val : array-like
            value in coordinate space
        method : str
            nearest, bounds, exact - each calculates the index differently
             - nearest : uses interp with coord values and rounds
             - bounds : uses interp between bounding values and truncates
             - exact : returns indices for exact coord values with other
                       indices masked (clean keyword has no effect)
        bounds : str
            ignore, error, warn if i,j are out of domain
        left : scalar
            see np.interp
        right : scalar
            see np.interp
        clean : {'none', 'mask'}
            none - return values regardless of bounds;
            mask - mask invalid values (use with left/right=np.nan);
                   has no affect with method exact

        Returns
        -------
        i : array-like
            indices (0-based) for variables
        """
        val = np.asarray(val)
        if method not in ('exact', 'nearest', 'bounds'):
            raise NotImplementedError(method + ' method not implemented')

        if bounds not in ('ignore', 'warn', 'error'):
            raise NotImplementedError(bounds + ' bounds not implemented')

        if clean not in ('none', 'mask'):
            raise NotImplementedError(clean + ' clean not implemented')

        dimv = self.variables[dim]
        dimvals = dimv[...]
        if dimvals.ndim > 1:
            raise ValueError(
                'val2idx is only implemented for 1-D coordinate variables'
            )
        if method == 'bounds':
            bounds_keys = [dim + '_bounds', dim + '_bnds']
            if hasattr(dimv, 'bounds'):
                bounds_keys.insert(0, dimv.bounds)

            for dimbk in bounds_keys:
                if dimbk not in self.variables:
                    continue
                dimbv = self.variables[dimbk]
                if dimbv.ndim == 1:
                    dimevals = dimbv[:]
                elif dimbv.ndim == 2 and dimbv.shape[1] == 2:
                    dimevals = np.append(dimbv[:, 0], dimbv[-1, 1])
                else:
                    raise ValueError(
                        'val2idx is only implemented for 1-D or 2-D bounds' +
                        '; {} has {}'.format(dimbk, dimbv.shape)
                    )
                break
            else:
                warn('Approximating bounds for val2idx {}'.format(dim))
                dval = np.diff(dimvals) / 2
                start = dimvals[:1]
                end = dimvals[-1:]
                if (dval == dval[0]).all():
                    start -= dval[0]
                    end += dval[-1]

                dimevals = np.concatenate([
                    start,
                    dimvals[1:] - dval,
                    end
                ])
        else:
            dimevals = dimvals

        idx = np.arange(dimevals.size)
        ddimevals = np.diff(dimevals)

        if (ddimevals < 0).all():
            dimevals[::-1]
            idx = idx[::-1]
        elif (ddimevals > 0).all():
            pass
        else:
            raise ValueError('coordinate is neither ascending nor descending')

        # import pdb; pdb.set_trace()
        # left = dimevals[0] - 1
        # right = dimevals[-1] + 1
        fidx = np.interp(val, dimevals, idx, left=left, right=right)
        if method == 'bounds':
            if right is None or right == dimevals[-1]:
                fidx = np.minimum(fidx, dimvals.size - 1)

        if method == 'exact':
            fidx = np.ma.masked_where(~np.in1d(val, dimvals), fidx)

        if clean == 'mask':
            outfidx = np.ma.masked_invalid(fidx)
        else:
            outfidx = fidx

        if bounds != 'ignore':
            isleft = val < dimevals[0]
            isright = val > dimevals[-1]
            isout = isleft | isright
            outmesg = 'Values are out of bounds:\n{}'.format(val[isout])
            if bounds == 'warn':
                warn(outmesg)
            elif bounds == 'error':
                raise ValueError(bounds)

        if method == 'nearest':
            outidx = np.round(outfidx, 0).astype('i')
        else:
            outidx = outfidx.astype('i')

        return outidx

    def get_dest(self):
        """
        Returns
        -------
        path : str
            path where a new file is created on some action

        Notes
        -----
        If None, a file is created in memory.
        Else, a netcdf file is created on disk.
        """
        return getattr(self, '_destination', None)

    def set_dest(self, path, **options):
        """
        Sets the path where a new file is created on some action

        Arguments
        ---------
        path : str
            path for new file
        **options : keywords for constructor
            options for new file creation

        Returns
        -------
        None
        """
        options['filename'] = path
        return object.__setattr__(self, '_destination', options)

    def get_varopt(self):
        """
        Get options

        Arguments
        ---------
        None

        Returns
        -------
        options : dictionary of options
        """
        return getattr(self, '_varopt', {}).copy()

    def set_varopt(self, **options):
        """
        Set options to be used when creating any Variable

        Arguments
        ---------
        **options : options for createVariable
            optional keywords to be supplied when creating new variables in
            destination file

        Returns
        -------
        None
        """
        return object.__setattr__(self, '_varopt', options)

    def _newlike(self):
        """
        Internal function to return a file of the same class if a
        PseudoNetCDFFile
        """
        if self.get_dest() is not None:
            outf = netcdf(**self.get_dest())
        elif isinstance(self, PseudoNetCDFFile):
            outt = type(self)
            outf = outt.__new__(outt)
        else:
            outf = PseudoNetCDFFile()
        outf.set_varopt(**self.get_varopt())
        return outf

    def renameVariable(self, oldkey, newkey, inplace=False, copyall=True):
        """
        Rename variable (oldkey)

        Parameters
        ----------
        oldkey : string
            variable to be renamed
        newkey : string
            new dame for variable
        inplace : boolean
            create the new variable in this netcdf file (default False)
        copyall : boolean
            if not inplace, should all variables be copied to new file

        Returns
        -------
        outf : PseudoNetCDFFile
            instance with renamed variable (this file if inplace = True)
        """
        return self.renameVariables(**{oldkey: newkey})

    def renameVariables(self, inplace=False, copyall=True, **newkeys):
        """
        Rename variables for each oldkey: newkey dictionary item

        Parameters
        ----------
        **newkeys : dictionary
            where key is the oldkey and value is the newkey
        inplace : boolean
            create the new variable in this netcdf file (default False)
        copyall :boolean
            if not inplace, should all variables be copied to new file

        Returns
        -------
        outf : PseudoNetCDFFile
            instance with renamed variable (this file if inplace = True)
        """
        if inplace:
            outf = self
        else:
            outf = self._copywith(
                props=True, dimensions=True, variables=copyall, data=copyall)

        for oldkey, newkey in newkeys.items():
            outf.copyVariable(self.variables[oldkey], key=newkey)
            if oldkey in outf.variables:
                del outf.variables[oldkey]

        return outf

    def renameDimension(self, oldkey, newkey, inplace=False):
        """
        Rename dimension (oldkey) in dimensions and in all variables

        Parameters
        ----------
        oldkey : string
            dimension to be renamed
        newkey : string
            new dame for dimension
        inplace : boolean
            create the new variable in this netcdf file (default False)

        Returns
        -------
        outf : PseudoNetCDFFile
            instance with renamed variable (this file if inplace = True)
        """
        return self.renameDimensions(inplace=inplace, **{oldkey: newkey})

    def renameDimensions(self, inplace=False, **newkeys):
        """
        Rename dimension (oldkey) in dimensions and in all variables

        Parameters
        ----------
        **newkeys : dictionary
            where key is the oldkey and value is the newkey
        inplace : boolean
            create the new variable in this netcdf file (default False)

        Returns
        -------
        outf : PseudoNetCDFFile
            instance with renamed variable (this file if inplace = True)
        """
        if inplace:
            outf = self
        else:
            outf = self.copy()

        for oldkey, newkey in newkeys.items():
            outf.dimensions[newkey] = outf.dimensions[oldkey]

        for k, v in outf.variables.items():
            olddims = v.dimensions
            newdims = tuple([newkeys.get(dk, dk) for dk in olddims])
            if newdims != olddims:
                v.dimensions = newdims
        for oldkey, newkey in newkeys.items():
            del outf.dimensions[oldkey]

        return outf

    def insertDimension(self, newonly=True, multionly=False, before=None,
                        after=None, inplace=False, **newdims):
        """
        Insert dimensions with keys and lengths from newdims


        Parameters
        ----------
        **newdims : dictionary
            where key is the new dimension and value is the length
        newonly : boolean
            Only add dimension to variables that do not already have it,
            default True
        multionly : boolean
            Only add dimension if there are already more than one (good
            for ignoring coordinate dimensions)
        before : string
            if variable has this dimension, insert the new dimension before
            it. Otherwise, add to the beginning. (before takes precedence)

        after : string
            if variable has this dimension, insert the new dimension after it.
            Otherwise, add to the beginning.

        inplace : boolean
            create the new variable in this netcdf file (default False)

        Returns
        -------
        outf : PseudoNetCDFFile
            instance will new dimension in dimensions and variables

        Notes
        -----

        1. Adding a non unity dimension will cause the data to be repeated
           along the new axis.
        2. If order of addition matters, use multiple calls. newdimsuse
           will be a non-ordered dictionary
        """
        if inplace:
            outf = self
        else:
            outf = self.copy(variables=False)

        for dk, dv in newdims.items():
            if dk not in outf.dimensions:
                outf.createDimension(dk, dv)

            for vk, vv in self.variables.items():
                vdims = list(vv.dimensions)
                if (
                    (newonly and dk in vdims) or
                    (multionly and len(vdims) == 1)
                ):
                    outf.copyVariable(vv, key=vk, withdata=True)
                    continue
                ndims = [_dk for _dk in vdims]
                if before in vdims:
                    bi = vdims.index(before)
                elif after in vdims:
                    bi = vdims.index(after) + 1
                elif not (before is None and after is None):
                    outf.copyVariable(vv, key=vk, withdata=True)
                    continue
                else:
                    bi = 0
                ndims.insert(bi, dk)
                var = outf.copyVariable(
                    vv, key=vk, dimensions=ndims, withdata=False)

                var[...] = np.expand_dims(vv[...], axis=bi)
        return outf

    def reorderDimensions(self, oldorder, neworder, inplace=False):
        """
        Evaluate expr and return a PseudoNetCDFFile object with resutl

        Parameters
        ----------
        oldorder : iterable of strings
            dimension names in existing order
        neworder : iterable of strings
            dimension names in new order

        Returns
        -------
        outf : PseudoNetCDFFile
            instance with dimensions reordered in variables
        """
        if inplace:
            outf = self
        else:
            outf = self.copy(variables=True)
        oldorder = tuple(oldorder)
        neworder = tuple(neworder)
        for vk, vv in self.variables.items():
            varneworder = [dk for dk in neworder if dk in vv.dimensions]
            varorder = [dk for dk in vv.dimensions]
            if len(varneworder) > 0:
                newvals = vv[:].copy()
                for newdi, newdk in enumerate(varneworder):
                    axisidx = varorder.index(newdk)
                    if axisidx == newdi:
                        continue
                    newvals = np.rollaxis(newvals, axis=axisidx, start=newdi)
                    varorder.pop(axisidx)
                    varorder.insert(newdi, newdk)
                assert(varorder == varneworder)
                newvals.dimensions = tuple(varorder)
                outf.variables[vk] = newvals
            else:
                pass

        return outf

    def eval(self, expr, inplace=False, copyall=False):
        """
        Evaluate expr and return a PseudoNetCDFFile object with resutl

        Parameters
        ----------
        expr : string
            expression to evaluate
        inplace : boolean
            create the new variable in this netcdf file (default False)
        copyall : boolean
            if not inplace, should all variables be copied to new file

        Returns
        -------
        outf : PseudoNetCDFFile
            instance with renamed variable (this file if inplace = True)
        """
        import numpy as np
        # Copy file to temporary PseudoNetCDF file
        comp = compile(expr, 'none', 'exec')
        vardict = {k: v for k, v in self.variables.items()}
        for pk in self.ncattrs():
            if pk not in vardict:
                vardict[pk] = getattr(self, pk)
        vardict['np'] = np
        vardict['self'] = self
        from symtable import symtable
        symtbl = symtable(expr, '<pncexpr>', 'exec')
        symbols = symtbl.get_symbols()
        for symbol in symbols:
            key = symbol.get_name()
            if key in vardict:
                tmpvar = vardict[key]
                if isinstance(tmpvar, (PseudoNetCDFVariable, NetCDFVariable)):
                    break
        else:
            key = 'N/A'
            tmpvar = PseudoNetCDFVariable(None, 'temp', 'f', ())

        if inplace:
            outf = self
        else:
            if copyall:
                outf = self.copy()
            else:
                newkeys = [key]
                outf = self.subsetVariables(newkeys)
                try:
                    del outf.variables[key]
                except Exception:
                    pass

        propd = dict([(k, getattr(tmpvar, k)) for k in tmpvar.ncattrs()])
        propd['expression'] = expr
        dimt = tmpvar.dimensions
        vardict['outf'] = self

        # Assign expression to new variable.
        exec(comp, None, vardict)
        assignedkeys = [s.get_name() for s in symbols if s.is_assigned()]
        assignedkeys = [k for k in assignedkeys if k in vardict]
        for key in assignedkeys:
            try:
                del outf.variables[key]
            except Exception:
                pass
            val = vardict[key]
            # if the output variable has no dimensions,
            # there is likely a problem and the output
            # should be defined.
            if (
                isinstance(val, (PseudoNetCDFVariable,)) and
                val.dimensions != ()
            ):
                outf.variables[key] = val
            else:
                outf.createVariable(key, val.dtype.char,
                                    dimt, values=val, **propd)

        return outf

    def mask(
        self, where=None, less=None, less_equal=None, greater=None,
        greater_equal=None, values=None, equal=None, invalid=False, mask=None,
        dims=None, fill_value=-999, coords=False, verbose=0
    ):
        """
        Apply mask to all variables of same shape or just where dimensions
        match.

        Parameters
        ----------
        where : array-like
            boolean array to use as a mask see numpy.ma.masked_where
        greater : scalar
            mask when values are greater than this value see
            numpy.ma.masked_greater
        less : scalar
            mask when values are less than this value see
            numpy.ma.masked_less
        greater_equal : scalar
            mask when values are greater than or equal to this value see
            numpy.ma.masked_greater
        less_equal : scalar
            mask when values are less than or equal to this value see
            numpy.ma.masked_less
        values : scalar
            mask when values are equal to this value within standard floating
            point see numpy.ma.masked_values
        equal : scalar
            mask when values are exactly this value (i.e., integers) see
            numpy.ma.masked_equal
        invalid : boolean
            mask when values are invalid, see numpy.ma.masked_invalid
        mask : array-like
            alias for where
        dims : iterable of strings
            only apply "mask" or "where" to variables with these dimensions
            no effet on other masks
        fill_value : scalar
            value to use as the fill_value for new arrays
        coords : boolean
            if True, apply masks to coordinate variables. Default, False
        verbose : int
            level of verbosity for function, mostly for debugging

        Returns
        -------
        outf : PseudoNetCDFFile
            instance with masked variables

        See Also
        --------
        numpy.ma : all masks are passing throught to numpy.ma.masked_...

        Notes
        -----
        mask options are not mutually exclusive. the order is where, greater,
        greater_equal, less, less_equal, values, equal, invalid
        """
        from time import time
        if where is None and mask is not None:
            where = mask
        elif mask is not None and where is not None:
            raise ValueError(
                'mask is an alias to where; supply one or the other, ' +
                'but not both'
            )

        if dims is None:
            maskdims = getattr(where, 'dimensions', dims)
        else:
            maskdims = dims

        coordkeys = self.getCoords()
        outf = self.copy(variables=False)
        nitems = len(self.variables.keys())
        if verbose == 1:
            print('|' + '=' * nitems + '|', flush=True)
            print('|', end='', flush=True)
        for vk, vv in self.variables.items():
            if verbose > 1:
                t0 = time()
                print(vk, end='', flush=True)
            elif verbose > 0:
                print('.', end='', flush=True)
            newvar = outf.copyVariable(
                vv, key=vk, fill_value=fill_value, withdata=False
            )
            if vk in coordkeys and not coords:
                newvar[...] = vv[...]
                continue

            vals = vv[...]
            if where is not None:
                if (
                    maskdims == vv.dimensions or
                    (
                        maskdims is None and
                        where.shape == vals.shape
                    )
                ):
                    vals = np.ma.masked_where(where, vals)
                    # np.ma.putmask(newvar[...], mask==False, vv[...])
                    # vals = newvar[...]
                # else:
                #     vals = vv[...]

            if greater is not None:
                vals = np.ma.masked_greater(vals, greater)

            if greater_equal is not None:
                vals = np.ma.masked_greater_equal(vals, greater_equal)

            if less is not None:
                vals = np.ma.masked_less(vals, less)

            if less_equal is not None:
                vals = np.ma.masked_less_equal(vals, less_equal)

            if values is not None:
                vals = np.ma.masked_values(vals, values)

            if equal is not None:
                vals = np.ma.masked_equal(vals, equal)

            if invalid:
                vals = np.ma.masked_invalid(vals)

            if verbose > 1:
                t1 = time()
                print(t1 - t0, end='\n', flush=True)

            newvar[...] = vals[...]
        if verbose > 1:
            pass
        elif verbose > 0:
            print('')
        return outf

    def plot(self, varkey, plottype=None, ax_kw=None,
             plot_kw=None, cbar_kw=None, map_kw=None, dimreduction='mean'):
        """
        Parameters
        ----------
        varkey : string
            the variable to plot
        plottype : string
            any dimension name pair delimited by a hyphen (e.g.,
            longitude-latitude, latitude-pressure, longitude-pressure,
            vertical-profile, time-longitude, time-latitude, time-pressure)
            defaults to the last two dimensions.
        ax_kw : dictionary
            keywords for the axes to be created
        plot_kw : dictionary
            keywords for the plot (plot, scatter, or pcolormesh) to be
            created
        cbar_kw : dictionary or bool or None
            keywords for the colorbar; if True or None, use defaults.
            If False, do not create a colorbar
        map_kw : dictionary or bool or None
            keywords for the getMap routine, which is only used with
            map capable dimensions (ie, plottype='longitude-latitude')
            If True or None, use default configuration dict(countries=True,
            coastlines=True, states=False, counties=False). If False,
            do not draw a map.
        dimreduction : string or function
            dimensions not being used in the plot are removed using
            applyAlongDimensions(dimkey=dimreduction) where each dimenions
        """
        import matplotlib.pyplot as plt
        from ..coordutil import getbounds

        if ax_kw is None:
            ax_kw = {}

        if plot_kw is None:
            plot_kw = {}

        if cbar_kw is None or cbar_kw is True:
            cbar_kw = {}

        if map_kw is None or map_kw is True:
            map_kw = {}
        elif map_kw is not False:
            map_kw = map_kw.copy()

        apply2dim = {}
        var = self.variables[varkey]
        if plottype is None:
            vdims = var.dimensions
            if len(vdims) > 1:
                plottype = '-'.join(var.dimensions[-2:][::-1])
            else:
                plottype = var.dimensions[0] + '-profile'

        varunit = varkey
        if hasattr(var, 'units'):
            varunit += ' ' + var.units.strip()

        dimlens = dict([(dk, len(self.dimensions[dk]))
                        for dk in var.dimensions])
        dimpos = dict([(dk, di) for di, dk in enumerate(var.dimensions)])
        xkey, ykey = plottype.split('-')
        if not ykey == 'profile':
            for dimkey in list(dimlens):
                if dimkey not in (xkey, ykey) and dimlens.get(dimkey, 1) > 1:
                    apply2dim[dimkey] = dimreduction

        if len(apply2dim) > 0:
            subsetkeys = [varkey]
            if xkey in self.variables:
                subsetkeys.append(xkey)
            if ykey in self.variables:
                subsetkeys.append(ykey)
            myf = self.subsetVariables(subsetkeys)\
                      .applyAlongDimensions(**apply2dim)
            var = myf.variables[varkey]
            dimlens = dict([(dk, len(self.dimensions[dk]))
                            for dk in var.dimensions])
        else:
            myf = self
        if ykey in ('profile',):
            vaxi = var.dimensions.index(xkey)
            vsize = var.shape[vaxi]
            vals = np.rollaxis(var[:], vaxi).reshape(vsize, -1)
        else:
            vals = var[:].squeeze()

        if xkey == 'time':
            xm = myf.getTimes()
            dx = np.diff(xm)[-1]
            x = np.append(xm, xm[-1] + dx)
            x = plt.matplotlib.dates.date2num(x)
        else:
            x = getbounds(myf, xkey)

        ax = plt.gca(**ax_kw)
        if ykey in ('profile',):
            y = getbounds(myf, xkey)
            x1 = vals[:].min(1)
            xm = vals[:].mean(1)
            x2 = vals[:].max(1)
            ax.fill_betweenx(y=y, x1=x1, x2=x2, label=varkey + '(min, max)')
            ax.plot(xm, y, label=varkey, **plot_kw)
            ax.set_ylabel(xkey)
            ax.set_xlabel(varunit)
            return ax

        if ykey == 'time':
            ym = myf.getTimes()
            dy = np.diff(ym)[-1]
            y = np.append(ym, ym[-1] + dy)
            y = plt.matplotlib.dates.date2num(y)
        else:
            y = getbounds(myf, ykey)

        if dimpos[xkey] < dimpos[ykey]:
            vals = vals.T
        p = ax.pcolormesh(x, y, vals, **plot_kw)
        if cbar_kw is not False:
            cbar_kw.setdefault('label', varunit)
            ax.figure.colorbar(p, **cbar_kw)
        if xkey == 'time':
            ax.xaxis.set_major_formatter(
                plt.matplotlib.dates.AutoDateFormatter(
                    plt.matplotlib.dates.AutoDateLocator()
                )
            )
        if ykey == 'time':
            ax.yaxis.set_major_formatter(
                plt.matplotlib.dates.AutoDateFormatter(
                    plt.matplotlib.dates.AutoDateLocator()
                )
            )
        mappabledims = (
            plottype == 'longitude-latitude' or
            plottype == 'lon-lat'
        )
        if mappabledims and map_kw is not False:
            try:
                map_kw = map_kw.copy()
                coastlines = map_kw.pop('coastlines', True)
                countries = map_kw.pop('countries', True)
                states = map_kw.pop('states', False)
                counties = map_kw.pop('counties', False)
                bmap = myf.getMap(**map_kw)
                if coastlines:
                    bmap.drawcoastlines(ax=ax)
                if countries:
                    bmap.drawcountries(ax=ax)
                if states:
                    bmap.drawstates(ax=ax)
                if counties:
                    bmap.drawcounties(ax=ax)
            except Exception:
                pass
        else:
            ax.set_xlabel(xkey)
            ax.set_ylabel(ykey)
        return ax

    def setncatts(self, attdict):
        """
        Set ncattrs from attdict keys and values

        Parameters
        ----------
        attdict : dictionary
            key/value pairs of properties

        Returns
        -------
        None
        """
        for pk, pv in attdict.items():
            setattr(self, pk, pv)

    def getncatts(self):
        """
        Return all ncattrs keys and values as a dictionary

        Returns
        ----------
        attdict : dictionary
            key/value pairs of properties
        """
        outd = OrderedDict()
        for pk in self.ncattrs():
            outd[pk] = getattr(self, pk)
        return outd

    @classmethod
    def from_ncvs(cls, *invars, **invarkw):
        """
        Arguments
        ---------
        invars : list
            NetCDF-like variable must have standard_name, long_name or name

        invarkw : kwds
            NetCDF-like variables

        Returns
        -------
        outf : PseudoNetcdf-like file
        """
        outf = cls()
        for invar in invars:
            for dk, ds in zip(invar.dimensions, invar.shape):
                if dk in outf.dimensions:
                    assert(ds == len(outf.dimensions[dk]))
                else:
                    outf.createDimension(dk, ds)
            outf.copyVariable(invar)

        for inkey, invar in invarkw.items():
            for dk, ds in zip(invar.dimensions, invar.shape):
                if dk in outf.dimensions:
                    assert(ds == len(outf.dimensions[dk]))
                else:
                    outf.createDimension(dk, ds)
            outf.copyVariable(invar, key=inkey)

        return outf

    @classmethod
    def from_ncf(cls, infile):
        """
        Arguments
        ---------
        infile : PseudoNetCDF-like file

        Returns
        -------
        outf : PseudoNetcdf-like file
        """
        outf = cls()
        for pk in infile.ncattrs():
            pv = getattr(infile, pk)
            setattr(outf, pk, pv)

        for dk, dv in infile.dimensions.items():
            outf.copyDimension(dv, key=dk)

        for vk, vv in infile.variables.items():
            outf.copyVariable(vv, key=vk)

        return outf

    def _copywith(self, props=True, dimensions=True, variables=False,
                  data=False):
        """
        Internal function for making copies of the same type

        Parameters
        ----------
        props : boolean
            include properties (default: True)
        dimensions : boolean
            include dimensions (default: True)
        variables : boolean
            include variable structures (default: False)
        data : boolean
            include variable data (default: False)

        Returns
        -------
        outf : PseudoNetCDFFile
            instance with some parts

        Notes
        -----
        Internal function does not return variables by default.
        This is useful for functions like slice, apply, eval, etc.

        The _ in _copywith means this is a private function and the
        call interface may change.
        """
        outf = self._newlike()
        outf._operator_exclude_vars = tuple(self._operator_exclude_vars)
        if props:
            for pk in self.ncattrs():
                setattr(outf, pk, getattr(self, pk))
        if dimensions:
            for dk, dv in self.dimensions.items():
                outf.copyDimension(dv, key=dk)

        if variables:
            for vk, vv in self.variables.items():
                outf.copyVariable(vv, key=vk, withdata=data)

        return outf

    def copy(self, props=True, dimensions=True, variables=True, data=True):
        """
        Function for making copies of the same type

        Parameters
        ----------
        props : boolean
            include properties (default: True)
        dimensions : boolean
            include dimensions (default: True)
        variables : boolean
            include variable structures (default: True)
        data : boolean
            include variable data (default: True)

        Returns
        -------
        outf : PseudoNetCDFFile instance
        """
        return self._copywith(props=props, dimensions=dimensions,
                              variables=variables, data=data)

    def interpDimension(self, dimkey, newdimvals, coordkey=None, **interpkwds):
        """
        Parameters
        ----------
        dimkey : string
            the new dimension for interpolation
        newdimvals : iterable
            the new values to interpolate to
        coordkey : string
            the variable to use as the old coordinate values
        interptype : string
            'linear' or 'conserve'. linear uses a linear interpolation
             conserve uses a mass conserving interpolation
        extrapolate : boolean
            allow extrapolation beyond bounds with linear, default False
        fill_value : numeric value
            set fill value (e.g, nan) to prevent extrapolation or edge
            continuation

        Returns
        -------
        outf : PseudoNetCDFFile
            instance with all variables interpolated

        Notes
        -----
        When extrapolate is false, the edge values are used for points beyond
        the inputs.
        """
        from ..coordutil import getinterpweights

        if coordkey is None:
            olddimvals = self.variables[dimkey]
        else:
            olddimvals = self.variables[coordkey]
        if olddimvals.ndim == 1 and newdimvals.ndim == 1:
            weights = getinterpweights(olddimvals, newdimvals, **interpkwds)

            def interpd(data):
                if data.ndim == 1:
                    newdata = (weights * data[:, None]).sum(0)
                else:
                    newdata = (weights[None, :, :, None, None] *
                               data[:, :, None]).sum(1)
                return newdata
            outf = self.applyAlongDimensions(**{dimkey: interpd})
        else:
            outf = self.copy(props=True, dimensions=False, variables=False)
            olddim = olddimvals.dimensions
            newdim = newdimvals.dimensions
            if olddim != newdim:
                raise ValueError('Can only interpolate if coordinate ' +
                                 'variable have the same named dimensions')
            dimaxis = olddim.index(dimkey)
            ndl = newdimvals.shape[dimaxis]
            for dk, dv in self.dimensions.items():
                if dk == dimkey:
                    dl = ndl
                else:
                    dl = len(dv)
                outf.copyDimension(dv, key=dk, dimlen=dl)

            for vk, vv in self.variables.items():
                nvv = outf.copyVariable(
                    vv, key=vk, withdata=vv.dimensions != newdim)

            Ni, Nk = olddimvals.shape[:dimaxis], olddimvals.shape[dimaxis + 1:]
            s_ = np.s_
            for ii in np.ndindex(Ni):
                for kk in np.ndindex(Nk):
                    od = olddimvals[ii + s_[:, ] + kk]
                    nd = newdimvals[ii + s_[:, ] + kk]
                    weights = getinterpweights(od, nd, **interpkwds)
                    for nvk, nvv in outf.variables.items():
                        if nvv.dimensions != newdim:
                            continue
                        vv = self.variables[nvk]
                        interpedv = (weights *
                                     vv[ii + s_[:, ] + kk][:, None]).sum(0)
                        nvv[ii + s_[..., ] + kk] = interpedv
        return outf

    def applyAlongDimensions(self, verbose=0, **dimfuncs):
        """
        Similar to numpy.apply_along_axis, but for damed dimensions and
        processes dimensions as well as variables

        Parameters
        ----------
        dimfuncs : dictionary
            key value pairs where the key is a dimensions and the value is a
            1D function (func1d) or a dictionary. If the value is a dictionary
            it must include func1d as a function and any keyword arguments as
            additional options
        verbose : integer
            0 silent, 1 show variable, 2 show dimensions and variables

        Returns
        -------
        outf : PseudoNetCDFFile
            instance with variables and dimensions after processing
        """
        outf = self._copywith(props=True, dimensions=False)
        dimlens = OrderedDict()
        for dk, dv in self.dimensions.items():
            dimlens[dk] = len(dv)

        for dk, df in dimfuncs.items():
            dv = self.dimensions[dk]
            if dk in dimfuncs:
                if verbose > 1:
                    print(dk, flush=True)
                if dk in self.variables:
                    dvar = self.variables[dk]
                    if dvar.ndim != 1:
                        dvar = np.arange(len(dv))
                else:
                    dvar = np.arange(len(dv))
                if isinstance(df, str):
                    newdl = getattr(dvar[...], df)(keepdims=True).size
                else:
                    newdl = df(dvar[:]).size
            else:
                newdl = len(dv)
            dimlens[dk] = newdl

        for dk, dv in self.dimensions.items():
            newdl = dimlens[dk]
            outf.copyDimension(dv, key=dk, dimlen=newdl)

        if verbose == 1:
            print('|' + '=' * len(self.variables.keys()) + '|', flush=True)
            print('|', end='', flush=True)

        for vark, varo in self.variables.items():
            vdims = varo.dimensions
            newvals = varo[...]
            dik = list(enumerate(vdims))
            for di, dk in dik[::-1]:
                if dk in dimfuncs:
                    if verbose > 1:
                        print(' ' * 100, '\r', vark, dk, end='\r', flush=True)
                    elif verbose > 0:
                        print('.', end='', flush=True)
                    opts = dict(axis=di, arr=newvals)
                    dfunc = dimfuncs[dk]
                    if isinstance(dfunc, dict):
                        opts.update(dfunc)
                        noopts = False
                    else:
                        opts['func1d'] = dfunc
                        noopts = True
                    if noopts and isinstance(dfunc, str):
                        newvals = getattr(newvals, dfunc)(
                            axis=di, keepdims=True)
                    else:
                        newvals = np.apply_along_axis(dfunc, di, newvals)
            newvaro = outf.copyVariable(varo, key=vark, withdata=False)
            newvaro[...] = newvals
        if verbose > 0:
            print()

        return outf

    def getTimes(self, datetype='datetime', bounds=False):
        """
        Get an array of datetime objects

        Parameters
        ----------
        datetype : string or numpy.dtype
            'datetime' or datetime64 dtype
        bounds : boolean
            get time boundaries

        Returns
        -------
        out : array
            datetime objects or array of numpy's datetype type

        Notes
        -----
        self must have a time or TFLAG variable
        """
        from PseudoNetCDF.coordutil import _parse_ref_date
        from datetime import date, datetime, timedelta, timezone
        utc = timezone.utc

        _calendaryearlike = {'noleap': 1970, '365_day': 1970,
                             'all_leap': 1972, '366_day': 1972}

        if 'time' in self.variables.keys():
            time = self.variables['time']
            timeunits = time.units.strip()
            calendar = getattr(time, 'calendar', 'gregorian').lower()
            if bounds:
                if 'time_bounds' in self.variables.keys():
                    time = self.variables['time_bounds']
                    time = np.append(time[:, 0], time[-1, 1])
                else:
                    dts = np.diff(time)
                    dt = dts.mean()
                    if (dt != dts).all():
                        warn('Time bounds are approximate')
                    time = np.append(time - dt / 2, time[-1] + dt / 2)
            if 'since' in timeunits:
                unit, base = timeunits.split(' since ')
                # Get the reference date
                refdate = _parse_ref_date(base)

                if calendar in _calendaryearlike:
                    refyear = refdate.year
                    # Get a year for relative day calculations
                    yearlike = _calendaryearlike[calendar]
                    # In that year, how many seconds and days are there
                    yearseconds = (date(yearlike + 1, 1, 1) -
                                   date(yearlike, 1, 1)).total_seconds()
                    yeardays = yearseconds / 3600 / 24

                    # Get a new reference date in yearlike
                    crefdate = datetime(yearlike, 1, 1, tzinfo=utc)
                    if refdate.month != 1 or refdate.day != 1:
                        # Get start date in yearlike
                        refcdate = datetime(
                            yearlike, refdate.month, refdate.day, tzinfo=utc)
                        # Calculate delta in years
                        addyears = (
                            crefdate - refcdate).total_seconds() / yearseconds
                    else:
                        addyears = 0
                    # Convert time to fractional years, including change in
                    # reference
                    incrdenom = {'years': 1, 'days': yeardays,
                                 'hours': yeardays * 24,
                                 'minutes': yeardays * 24 * 60,
                                 'seconds': yeardays * 24 * 60}[unit]
                    fracyearincrs = time[:] / incrdenom + addyears
                    # Split into years and days
                    yearincrs = np.array(fracyearincrs // 1).astype('i')
                    dayincrs = (fracyearincrs % 1) * yeardays
                    # Add days to the calendar year reference
                    cdays = [crefdate + timedelta(days=dayinc)
                             for dayinc in dayincrs]
                    try:
                        # Combine calendar specific month and day with new year
                        out = np.array([
                            datetime(refyear + yearinc, cday.month,
                                     cday.day, tzinfo=utc)
                            for yearinc, cday in zip(yearincrs, cdays)])
                    except Exception:
                        warn(('Years calculated from %d day year, but ' +
                              'month/days calculated for actual year. ' +
                              'Usually means data has Feb 29th in a non ' +
                              'leap year') % yeardays)
                        out = np.array([
                            datetime(refyear + yearinc, 1, 1, tzinfo=utc) +
                            timedelta(days=float(dayinc))
                            for yearinc, dayinc in zip(yearincrs, dayincrs)])

                else:
                    out = refdate + \
                        np.array([timedelta(**{unit: float(i)})
                                  for i in time[:]])
            else:
                return time
        elif 'TFLAG' in self.variables.keys():
            dates = self.variables['TFLAG'][:][:, 0, 0]
            if (dates == -635).any():
                warn('Dates of -635 set to 1970001')
                dates[dates == -635] = 1970001
            times = self.variables['TFLAG'][:][:, 0, 1]
            yyyys = (dates // 1000).astype('i')
            jjj = dates % 1000
            hours = times // 10000
            minutes = times % 10000 // 100
            seconds = times % 100
            days = jjj + (hours + minutes / 60. + seconds / 3600.) / 24.
            out = np.array([datetime(yyyy, 1, 1, tzinfo=utc) +
                            timedelta(days=day - 1)
                            for yyyy, day in zip(yyyys, days)])
            if bounds:
                if hasattr(self, 'TSTEP'):
                    tstep = getattr(self, 'TSTEP')
                    sh = tstep // 10000 * 3600
                    sm = tstep % 10000 // 100 * 60
                    ss = tstep % 100
                    dt = timedelta(seconds=sh + sm + ss)
                else:
                    dts = np.diff(out)
                    dt = dts.mean()
                    if (dt != dts).all():
                        warn('Time bounds are approximate')
                out = np.append(out, out[-1] + dt)
        elif hasattr(self, 'SDATE') and hasattr(self, 'STIME') and \
                hasattr(self, 'TSTEP') and 'TSTEP' in self.dimensions:
            jdate = self.SDATE
            if jdate < 1:
                warn(f'SDATE was {jdate}; using 1970001')
                jdate = 1970001
            hhmmss = self.STIME
            refdate = datetime.strptime(
                '%07d %06d+0000' % (jdate, hhmmss), '%Y%j %H%M%S%z')
            tstepstr = '%06d' % self.TSTEP
            timeincr = timedelta(seconds=int(tstepstr[-2:])) + \
                timedelta(minutes=int(tstepstr[-4:-2])) + \
                timedelta(hours=int(tstepstr[:-4]))
            ntimes = len(self.dimensions['TSTEP'])
            if bounds:
                ntimes += 1
            timeincrs = timeincr * np.arange(ntimes)
            out = refdate + timeincrs
        elif 'tau0' in self.variables.keys():
            out = datetime(1985, 1, 1, 0, tzinfo=utc) + \
                np.array([timedelta(hours=i)
                          for i in self.variables['tau0'][:]])
            if bounds and 'tau1' in self.variables.keys():
                oute = (
                    datetime(1985, 1, 1, 0, tzinfo=utc) +
                    np.array([
                        timedelta(hours=i)
                        for i in self.variables['tau1'][:]
                    ])
                )
                out = np.append(out, oute[-1])
        else:
            raise ValueError('cannot understand time for file')
        if datetype == 'datetime':
            return out
        else:
            return np.array(out, dtype=datetype)

    def stack(self, other, stackdim):
        """
        Concatenates all variables on stackdim

        Parameters
        ----------
        other : instance or list of PseudoNetCDFFiles
            files to add to this file along stackdim

        stackdim : str
            dimension name

        Returns
        -------
        outf : PseudoNetCDFFile
            instance with stacked variables and dimension equal to new length
        """
        from collections.abc import Iterable
        outf = self._copywith(props=True, dimensions=False)
        if isinstance(other, Iterable):
            fs = [self] + list(other)
        else:
            fs = [self, other]
        dimensions = [f_.dimensions for f_ in fs]
        shareddims = {}
        for dimk, dim in self.dimensions.items():
            if dimk == stackdim:
                continue
            dimlens = [len(dims[dimk]) for dims in dimensions]
            if all([len(dim) == i for i in dimlens]):
                shareddims[dimk] = len(dim)
        differentdims = [set(dims.keys()).difference(
            shareddims.keys()) for dims in dimensions]
        assert(all([different.union([stackdim]) == set([stackdim])
                    for different in differentdims]))
        for dimkey in shareddims:
            ind = self.dimensions[dimkey]
            outf.copyDimension(ind, key=dimkey)
        newdl = sum([len(dims[stackdim]) for dims in dimensions])
        outf.copyDimension(self.dimensions[stackdim],
                           key=stackdim, dimlen=newdl)
        for tmpf in fs:
            for varkey, var in tmpf.variables.items():
                if stackdim not in var.dimensions:
                    if varkey in outf.variables:
                        if (
                            np.array_equal(outf.variables[varkey][...],
                                           var[...])
                        ):
                            pass
                        elif varkey not in self.dimensions:
                            warn(('Got duplicate variables for %s without ' +
                                  'stackable dimension; first value retained')
                                 % varkey)
                        continue
                    else:
                        outvals = var[...]
                else:
                    if varkey not in outf.variables.keys():
                        axisi = list(var.dimensions).index(stackdim)
                        outvals = np.ma.concatenate(
                            [f_.variables[varkey][:] for f_ in fs], axis=axisi)
                    else:
                        continue
                outvar = outf.copyVariable(var, key=varkey, withdata=False)
                outvar[...] = outvals

        return outf

    def subsetVariables(
        self, varkeys, inplace=False, exclude=False, keepcoords=True
    ):
        """
        Return a PseudoNetCDFFile with only varkeys

        Parameters
        ----------
        varkeys : iterable of strings
            keys to keep
        inplace : boolean
            if true (default false), then remove other variable from
            this file
        exclude : boolean
            if True (default False), then remove just these variables
        keepcoords : boolean
            if True (default True), keep coordinate variables

        Returns
        -------
        outf : PseudoNetCDFFile
            instance with variables
        """
        if exclude:
            varkeys = list(set(list(self.variables)).difference(varkeys))

        varkeys = varkeys + [k for k in self.getCoords() if k not in varkeys]

        if inplace:
            outf = self
            for varkey in list(outf.variables):
                if varkey not in varkeys:
                    del outf.variables[varkey]
        else:
            outf = self._copywith(props=True, dimensions=True)
            for varkey in varkeys:
                varo = self.variables[varkey]
                outf.copyVariable(varo, key=varkey, withdata=True)
        return outf

    def sliceDimensions(self, newdims=('POINTS',), verbose=0, **dimslices):
        """
        Return a netcdflike object with dimensions sliced

        Parameters
        ----------
        newdims : iterable of strings
            names for new dimensions. When more than one iterable applies to a
            variable slice, fancy indexing removes both dimensions and creates
            a new one of the iterable lengths
        **dimslices : dictionary
            key value pairs where the key is a dimension and the value is a
            valid slice object (slices, ints or iterables) if iterables are
            provided, all iterables must be the same size and shape. If the
            arrays are not 1D, newdims must have ndim names

        Returns
        -------
        outf : PseudoNetCDFFile
            instance with variables and dimensions sliced
        """
        outf = self._copywith(props=True, dimensions=False)
        newdimlens = OrderedDict()
        for dk, dv in self.dimensions.items():
            newdimlens[dk] = len(dv)

        isarray = {dk: not np.isscalar(dv) and not isinstance(
            dv, slice) for dk, dv in dimslices.items()}
        anyisarray = np.sum(list(isarray.values())) > 1

        if anyisarray:
            arraylens = np.array(
                [np.asarray(da).size
                 for dk, da in dimslices.items() if isarray[dk]])
            arrayshapes = np.array(
                [np.asarray(da).shape
                 for dk, da in dimslices.items() if isarray[dk]])
            arraylen = arraylens[0]
            arrayshape = arrayshapes[0]
            if (
                not (arraylens == arraylen).all() or
                not (arrayshapes == arrayshape).all()
            ):
                raise ValueError(
                    'If slicing with arrays, they must all be the same size ' +
                    'and shape')
            for dk, ia in isarray.items():
                if ia:
                    dimslices[dk] = np.asarray(dimslices[dk])

        for dk, ds in dimslices.items():
            # if anyisarray and isarray[dk]: continue
            dv = self.dimensions[dk]
            if dk in dimslices:
                if dk in self.variables:
                    dvar = self.variables[dk]
                else:
                    dvar = np.arange(len(dv))
                newdl = dvar[dimslices[dk]].size
            else:
                newdl = len(dv)
            newdimlens[dk] = newdl

        for dk, dl in newdimlens.items():
            dv = self.dimensions[dk]
            outf.copyDimension(dv, key=dk, dimlen=dl)

        if anyisarray:
            for ni, newdim in enumerate(newdims):
                outf.createDimension(newdim, arrayshape[ni])

        if verbose == 1:
            print('|' + '=' * len(self.variables.keys()) + '|', flush=True)
            print('|', end='', flush=True)
        for vark, varo in self.variables.items():
            if verbose > 1:
                print(' ' * 100, '\r', vark, end='\r', flush=True)
            elif verbose > 0:
                print('.', end='', flush=True)
            odims = vdims = varo.dimensions
            sliceo = tuple(dimslices.get(dk, slice(None)) for dk in vdims)
            isdarray = [isarray.get(dk, False) for dk in vdims]
            needsfancy = sum(isdarray) > 1
            if anyisarray and needsfancy:
                concatax = np.argmax(isdarray)
                odims = [dk for dk in vdims if not isarray.get(dk, False)]
                for newdim in newdims[::-1]:
                    odims.insert(concatax, newdim)

            newvaro = outf.copyVariable(
                varo, key=vark, dimensions=odims, withdata=False)
            for pk in varo.ncattrs():
                setattr(newvaro, pk, getattr(varo, pk))
            if anyisarray and needsfancy:
                point_arrays = []
                for ii in range(arraylen):
                    sliceoi = []
                    for si in sliceo:
                        if np.isscalar(si):
                            sliceoi.append([si])
                        elif isinstance(si, slice):
                            sliceoi.append(si)
                        else:
                            sliceoi.append(si.ravel()[ii])
                    sliceoi = tuple(sliceoi)
                    point_arrays.append(np.expand_dims(
                        varo[sliceoi], axis=concatax))
                newvals = np.concatenate(point_arrays, axis=concatax)
            else:
                newvals = varo[sliceo]
            try:
                newvaro[...] = newvals
            except Exception:
                newvaro[...] = newvals.reshape(newvaro.shape)

        if verbose > 0:
            print()
        return outf

    def removeSingleton(self, dimkey=None):
        """
        Return a netcdflike object with dimensions sliced

        Parameters
        ----------
        dimkey : string
            key of dimension to be evaluated for removal; if None, evaluate
            all. only singleton dimensions will be removed.

        Returns
        -------
        outf : PseudoNetCDFFile
            instance with dimensions removed
        """
        outf = self._copywith(props=True, dimensions=False)
        removed_dims = []
        for dk, d in self.dimensions.items():
            ni = len(d)
            if (dimkey is None or dk == dimkey) and ni == 1:
                removed_dims.append(dk)
            else:
                outf.copyDimension(d, key=dk, dimlen=ni)

        for vk, v in self.variables.items():
            olddims = v.dimensions
            newdims = tuple(
                [dk for dk in v.dimensions if dk not in removed_dims])
            sdims = tuple([(di, dk) for di, dk in enumerate(
                olddims) if dk not in newdims])[::-1]
            propd = dict([(pk, getattr(v, pk)) for pk in v.ncattrs()])
            ov = outf.createVariable(vk, v.dtype.char, newdims, **propd)
            outvals = v[...]
            for di, dk in sdims:
                outvals = outvals.take(0, axis=di)

            ov[...] = outvals[...]
        return outf

    def __repr__(self):
        from PseudoNetCDF.pncdump import pncdump
        import sys
        if sys.version_info.major == 3:
            from io import StringIO
        else:
            from StringIO import StringIO

        out = StringIO()
        pncdump(self, header=True, outfile=out)
        out.seek(0, 0)
        return out.read()

    @classmethod
    def isMine(cls, *args, **kwds):
        """
        True if this file or object can be identified
        for use by this class. Useful to override for
        classes that can be initialized from disk.
        """
        return False

    def __new__(mcl, *args, **kwds):
        new = super(PseudoNetCDFFile, mcl).__new__(mcl)
        new.variables = OrderedDict()
        new.dimensions = PseudoNetCDFDimensions()
        new._ncattrs = ()
        new._operator_exclude_vars = ()
        return new

    def __init__(self, *args, **properties):
        mode = properties.pop('mode', 'w')
        self._mode = mode
        for k, v in properties.items():
            setattr(self, k, v)

    @classmethod
    def open_mfdataset(cls, *paths, stackdim=None, **kwds):
        files = [cls(p, **kwds) for p in paths]
        file1 = files[0]
        if stackdim is None:
            for dk, dv in file1.dimensions.items():
                if dv.isunlimited():
                    stackdim = dk
            else:
                for dk in list(file1.dimensions):
                    if dk in ('TSTEP', 'time', 'Time', 't'):
                        stackdim = dk
                else:
                    raise ValueError(
                        'No dimension is unlimited or time; ' +
                        'must specify stackdim'
                    )
        return file1.stack(files[1:], stackdim=stackdim)

    def iswritable(self):
        return (self._mode[:1] in ('a', 'w') or self._mode[:2] in ('r+',))

    def __setattr__(self, k, v):
        if not (k[:1] == '_' or k in ('dimensions', 'variables', 'groups')):
            if k not in self._ncattrs:
                self._ncattrs += (k, )
        object.__setattr__(self, k, v)

    def __delattr__(self, k):
        if k in self._ncattrs:
            self._ncattrs = tuple([k_ for k_ in self._ncattrs if k_ != k])
        object.__delattr__(self, k)

    def setCoords(self, keys, missing='ignore'):
        """
        Set a variable as a coordinate variable

        Parameters
        ----------
        keys : iterable of strings
            keys for coord variables
        missing : string
            action if missing 'ignore', 'skip' or 'error'
            ignore - add in case used later
            skip   - do not add
            error  - raise an error

        Returns
        -------
        None

        Notes
        -----
        Coordinate variables are excluded from math
        """

        if missing == 'ignore':
            pass
        elif missing == 'skip':
            keys = [key for key in keys if key in self.variables]
        elif missing == 'error':
            invalid_keys = [key for key in keys if key not in self.variables]
            if len(invalid_keys) > 0:
                raise KeyError('File does not contain the variables: ' +
                               '{}'.format(invalid_keys))
        else:
            raise ValueError(
                'Must be ignore, skip or error; received {}'.format(missing))

        new = tuple(set(self._operator_exclude_vars).union(set(keys)))
        try:
            self._operator_exclude_vars = new
        except Exception:
            return new

    def getCoords(self):
        """
        Return a list of coordkeys
        """
        return tuple([k for k in self._operator_exclude_vars])

    def createDimension(self, name, length):
        """
        Create a dimension

        Parameters
        ----------
        name : string
            name for dimension
        length : integer
            maximum length of dimension

        Returns
        -------
        dim : PseudoNetCDFDimensions
            new dimension
        """
        dim = self.dimensions[name] = PseudoNetCDFDimension(self, name, length)
        return dim

    def copyDimension(self, dim, key=None, dimlen=None, unlimited=None):
        if dimlen is None:
            dimlen = len(dim)

        if key is None:
            key = dim.name

        if unlimited is None:
            unlimited = dim.isunlimited()

        if isinstance(self, netcdf):
            if unlimited:
                ndv = self.createDimension(key, None)
            else:
                ndv = self.createDimension(key, dimlen)
        else:
            ndv = self.createDimension(key, dimlen)
            ndv.setunlimited(unlimited)

        return ndv

    def copyVariable(self, var, key=None, dtype=None, dimensions=None,
                     fill_value=None, withdata=True):
        """
        Copy var into self as vark

        Parameters
        ----------
        var : PseudoNetCDFVariable
            netCDF4.Variable-like object (must have ncattrs and setncatts)
        key : string
            key for variable in self (can be omitted if var has name,
            standard_name, or long_name)
        dtype : string or numpy.dtype
            change the data type to dtype
        dimensions : iterable of strings
            change the dimensions to dimensions
        fill_value : integer or flaot
            change the fill_value to this values
        withdata : boolean
            default True, copies data

        Returns
        -------
        myvar : PseuodNetCDFVairable
            copy of var in this file
        """
        if key is None:
            for propk in ['_name', 'name', 'standard_name', 'long_name']:
                if hasattr(var, propk):
                    key = getattr(var, propk)
                    break
            else:
                raise AttributeError(
                    'varkey must be supplied because var has no name, ' +
                    'standard_name or long_name')

        if withdata:
            try:
                vals = var[:]
            except Exception:
                vals = var[...]
        else:
            vals = var

        if dtype is None:
            dtype = vals.dtype

        if dimensions is None:
            dimensions = var.dimensions
        if fill_value is None:
            for pk in ('fill_value', 'missing_value', '_FillValue'):
                fill_value = getattr(var, pk, None)
                if fill_value is not None:
                    break

        myvar = self.createVariable(
            key, dtype, dimensions, fill_value=fill_value)
        attrs = OrderedDict()
        for propk in var.ncattrs():
            attrs[propk] = getattr(var, propk)
        myvar.setncatts(attrs)
        if withdata:
            try:
                myvar[:] = vals[:]
            except Exception:
                myvar[...] = vals[...]
        return myvar

    def createVariable(self, name, type, dimensions, fill_value=None,
                       **properties):
        """
        Create a variable

        Parameters
        ----------
        name : string
            name for new variable
        type : string or numpy dtype
            code (e.g., 'f', 'i', 'd')
        dimensions : tuple of strigns
            dimension keys that can be found in objects' dimensions dictionary

        Returns
        -------
        var : new variable
        """
        import numpy as np
        varopt = self.get_varopt()
        if fill_value is None:
            fill_value = varopt.get('fill_value', None)

        for pk, pv in varopt.items():
            if pk != 'fill_value':
                properties.setdefault(pk, pv)

        if fill_value is None:
            for pk in 'missing_value _FillValue'.split():
                fill_value = properties.get(pk, None)
                if fill_value is not None:
                    break
        else:
            properties['fill_value'] = fill_value

        if type == 'S':
            type = 'c'
        if (
            isinstance(properties.get('values', 1), np.ma.MaskedArray) or
            fill_value is not None
        ):
            var = self.variables[name] = PseudoNetCDFMaskedVariable(
                self, name, type, dimensions, **properties)
        else:
            var = self.variables[name] = PseudoNetCDFVariable(
                self, name, type, dimensions, **properties)
        return var

    def close(self):
        """
        Does nothing.  Implemented for continuity with Scientific.IO.NetCDF
        """
        pass

    def dump(self, *args, **kwds):
        from PseudoNetCDF.pncdump import pncdump
        pncdump(self, *args, **kwds)

    def save(self, *args, **kwds):
        """
        Provides access to pncwrite for self

        Parameters
        ----------
        see Help pncwrite

        Returns
        -------
        see Help pncwrite
        """
        from PseudoNetCDF import pncwrite
        return pncwrite(self, *args, **kwds)

    def ncattrs(self):
        return self._ncattrs

    def setncattr(self, k, v):
        return setattr(self, k, v)

    def delncattr(self, k):
        self.__delattr__(k)

    def __add__(self, lhs):
        from ._functions import pncbo
        return pncbo(op='+', ifile1=self, ifile2=lhs, verbose=0,
                     coordkeys=self._operator_exclude_vars)

    def __sub__(self, lhs):
        from ._functions import pncbo
        return pncbo(op='-', ifile1=self, ifile2=lhs, verbose=0,
                     coordkeys=self._operator_exclude_vars)

    def __mul__(self, lhs):
        from ._functions import pncbo
        return pncbo(op='*', ifile1=self, ifile2=lhs, verbose=0,
                     coordkeys=self._operator_exclude_vars)

    def __truediv__(self, lhs):
        from ._functions import pncbo
        return pncbo(op='/', ifile1=self, ifile2=lhs, verbose=0,
                     coordkeys=self._operator_exclude_vars)

    def __floordiv__(self, lhs):
        from ._functions import pncbo
        return pncbo(op='//', ifile1=self, ifile2=lhs, verbose=0,
                     coordkeys=self._operator_exclude_vars)

    def __pow__(self, lhs):
        from ._functions import pncbo
        return pncbo(op='**', ifile1=self, ifile2=lhs, verbose=0,
                     coordkeys=self._operator_exclude_vars)

    def __and__(self, lhs):
        from ._functions import pncbo
        return pncbo(op='&', ifile1=self, ifile2=lhs, verbose=0,
                     coordkeys=self._operator_exclude_vars)

    def __or__(self, lhs):
        from ._functions import pncbo
        return pncbo(op='|', ifile1=self, ifile2=lhs, verbose=0,
                     coordkeys=self._operator_exclude_vars)

    def __xor__(self, lhs):
        from ._functions import pncbo
        return pncbo(op='^', ifile1=self, ifile2=lhs, verbose=0,
                     coordkeys=self._operator_exclude_vars)

    def __mod__(self, lhs):
        from ._functions import pncbo
        return pncbo(op='%', ifile1=self, ifile2=lhs, verbose=0,
                     coordkeys=self._operator_exclude_vars)

    def __lt__(self, lhs):
        from ._functions import pncbo
        return pncbo(op='<', ifile1=self, ifile2=lhs, verbose=0,
                     coordkeys=self._operator_exclude_vars)

    def __gt__(self, lhs):
        from ._functions import pncbo
        return pncbo(op='>', ifile1=self, ifile2=lhs, verbose=0,
                     coordkeys=self._operator_exclude_vars)

    def __eq__(self, lhs):
        if isinstance(lhs, (NetCDFFile, PseudoNetCDFFile)):
            from ._functions import pncbo
            return pncbo(op=' == ', ifile1=self, ifile2=lhs, verbose=0,
                         coordkeys=self._operator_exclude_vars)
        else:
            return lhs.__eq__(self)

    def __le__(self, lhs):
        from ._functions import pncbo
        return pncbo(op='<=', ifile1=self, ifile2=lhs, verbose=0,
                     coordkeys=self._operator_exclude_vars)

    def __ge__(self, lhs):
        from ._functions import pncbo
        return pncbo(op='>=', ifile1=self, ifile2=lhs, verbose=0,
                     coordkeys=self._operator_exclude_vars)

    def __ne__(self, lhs):
        from ._functions import pncbo
        return pncbo(op='!=', ifile1=self, ifile2=lhs, verbose=0,
                     coordkeys=self._operator_exclude_vars)

    slice = sliceDimensions
    apply = applyAlongDimensions
    subset = subsetVariables
    sync = close
    flush = close


_ncvarkwds = [
    'varname', 'datatype', 'dimensions', 'zlib', 'complevel', 'shuffle',
    'fletcher32', 'contiguous', 'chunksizes', 'endian',
    'least_significant_digit', 'fill_value'
]


class netcdf(PseudoNetCDFFile, NetCDFFile):
    def createDimension(self, *args, **kwds):
        return NetCDFFile.createDimension(self, *args, **kwds)

    def createVariable(self, *args, **kwds):
        for pk, pv in self.get_varopt().items():
            kwds.setdefault(pk, pv)
        propkwds = OrderedDict()
        for pk in set(kwds).difference(_ncvarkwds):
            propkwds[pk] = kwds.pop(pk)

        outval = NetCDFFile.createVariable(self, *args, **kwds)
        for pk, pv in propkwds.items():
            outval.setncattr(pk, pv)

        return outval

    def setncattr(self, k, v):
        return NetCDFFile.setncattr(self, k, v)

    def __setattr__(self, k, v):
        return NetCDFFile.__setattr__(self, k, v)

    def __delattr__(self, k):
        NetCDFFile.__delattr__(self, k)

    def __new__(cls, *args, **kwds):
        return NetCDFFile.__new__(cls, *args, **kwds)

    def setCoords(self, keys, missing='ignore'):
        new = PseudoNetCDFFile.setCoords(self, keys, missing=missing)
        self.__dict__['_operator_exclude_vars'] = tuple(set(new))

    def __init__(self, *args, **kwds):
        NetCDFFile.__init__(self, *args, **kwds)
        self.__dict__['_mode'] = kwds.get('mode', 'r')
        self.__dict__['_operator_exclude_vars'] = ()
        self.__dict__['_varopt'] = {}
        coords = [k for k in self.dimensions if k in self.variables]
        self.setCoords(coords)

    @property
    def _mode(self):
        return self.__dict__.get('_mode', 'r')

    def get_dest(self):
        """
        Returns the path where a new file is created on some action

        If None, a file is created in memory.
        Else, a netcdf file is created on disk.
        """
        return self.__dict__.get('_destination', None)

    def set_dest(self, path, **options):
        """
        Sets the path where a new file is created on some action

        Arguments
        ---------
        path : path for new file
        options : options for new file creation

        Returns
        -------
        None
        """
        options['filename'] = path
        self.__dict__['_destination'] = options

    def get_varopt(self):
        """
        Get options

        Arguments
        ---------
        None

        Returns
        -------
        options : dictionary of optiosn
        """
        return self.__dict__.get('_varopt', {}).copy()

    def set_varopt(self, **options):
        """
        Set options to be used when creating any Variable

        Arguments
        ---------
        options : options for new Variable creation

        Returns
        -------
        None
        """
        self.__dict__['_varopt'] = options

    def ncattrs(self):
        return NetCDFFile.ncattrs(self)

    @classmethod
    def from_ncf(cls, infile):
        outf = PseudoNetCDFFile()
        for pk in infile.ncattrs():
            pv = getattr(infile, pk)
            setattr(outf, pk, pv)

        for dk, dv in infile.dimensions.items():
            outf.copyDimension(dv, key=dk)

        for vk, vv in infile.variables.items():
            outf.copyVariable(vv, key=vk)

        return outf

    def _newlike(self):
        """
        Internal function to return a file of the same class if a
        PsueoNetCDFFile
        """
        if self.get_dest() is not None:
            outf = netcdf(**self.get_dest())
        else:
            outf = PseudoNetCDFFile()
        outf.set_varopt(**self.get_varopt())
        return outf

    @classmethod
    def isMine(cls, path, *args, **kwds):
        """
        True if this file or object can be identified
        for use by this class. Useful to override for
        classes that can be initialized from disk.
        """
        tmpf = open(path, 'rb')
        tmpf.seek(0, 0)
        cdftest = tmpf.read(3)
        tmpf.seek(1, 0)
        hdftest = tmpf.read(3)
        tmpf.close()
        if cdftest == b'CDF':
            return True
        elif hdftest == b'HDF':
            try:
                cls(path, *args, **kwds)
                return True
            except Exception:
                return False
        else:
            return False

    def close(self):
        try:
            return NetCDFFile.close(self)
        except Exception as e:
            warn(str(e))

    def __del__(self):
        self.close()


registerreader('nc', netcdf)
registerreader('ncf', netcdf)


class PseudoNetCDFVariables(OrderedDefaultDict):
    """
    PseudoNetCDFVariables provides a special implementation
    of the default dictionary that provides efficient access
    to variables of a PseudoNetCDFFile.  PseudoNetCDFFiles may
    have large variables that should only be loaded if accessed.
    PseudoNetCDFVariables allows a user to specify a function
    that can create variables on demand.
    """

    def __init__(self, func, keys):
        """
        func: function
            must take a key and provides a PseudoNetCDFVariable
        keys: iterable of strings
            keys that the dictionary should act as if it has
        """
        super(PseudoNetCDFVariables, self).__init__()
        self.__func = func
        self.__keys = keys

    def __missing__(self, k):
        """
        If the dictionary does not have a key, check if the
        user has provided that key.  If so, call the user
        specifie function to create the variable.
        """
        if k in self.keys():
            return self.__func(k)
        else:
            raise KeyError('missing "%s"' % (k, ))

    def addkey(self, k):
        """
        Allow the user to extend keys after the object
        has been created.
        """
        if k not in self.__keys:
            self.__keys.append(k)

    def keys(self):
        return tuple(
            self.__keys +
            [k for k in dict.keys(self) if k not in self.__keys])

    def __iter__(self):
        for k in self.keys():
            yield k

    def __len__(self):
        return len([k for k in self.keys()])

    def items(self):
        return [(k, self[k]) for k in self.keys()]

    def __contains__(self, k):
        return k in [k for k in self.keys()]


if __name__ == '__main__':
    unittest.main()
